import pytest

from psautohint import _psautohint


INFO = b"FontName Foo"
NAME = b"Foo"
GLYPH = b"""% square
0 500 rb
60 500 ry
sc
560 500 mt
560 0 dt
60 0 dt
60 500 dt
cp
ed
"""


def test_autohint_good_args():
    _psautohint.autohint(INFO, GLYPH)


def test_autohintmm_good_args():
    _psautohint.autohintmm((GLYPH, GLYPH), (NAME, NAME))


@pytest.mark.parametrize("args", [
    [],                               # no arguments
    [INFO],                           # 1 argument
    [INFO.decode('ascii'), GLYPH],    # 1st is string not bytes
    [INFO, GLYPH.decode('ascii')],    # 2nd is string not bytes
    [[INFO], GLYPH.decode('ascii')],  # 1st is a list
    [INFO, [GLYPH.decode('ascii')]],  # 2nd is a list
])
def test_autohint_bad_args(args):
    with pytest.raises(TypeError):
        _psautohint.autohint(*args)


@pytest.mark.parametrize("args", [
    [],                          # no arguments
    [(GLYPH, GLYPH)],            # 1 argument
    [GLYPH, (NAME, NAME)],       # 1st is not a tuple
    [(GLYPH, GLYPH), NAME],      # 2nd is not a tuple
    [(GLYPH, GLYPH), tuple()],   # 2nd is an empty tuple
    [(GLYPH, GLYPH), (NAME,)],   # 2nd is shorter than 1st
    [(GLYPH,), (NAME,)],         # 1st is one glyph
])
def test_autohintmm_bad_args(args):
    with pytest.raises(TypeError):
        _psautohint.autohintmm(*args)


@pytest.mark.parametrize("args", [
    [(GLYPH.decode('ascii'), GLYPH), (NAME, NAME)],  # 1st should be bytes
    [(GLYPH, GLYPH), (NAME.decode('ascii'), NAME)],  # 2nd should be bytes
])
def test_autohintmm_unicode(args):
    with pytest.raises(TypeError):
        _psautohint.autohintmm(*args)


@pytest.mark.parametrize("glyph", [
    b"% foo\ned",        # ending comment with newline
    b"% foo\red",        # ending comment with linefeed
    b"\t% foo\nsc\ted",  # separating tokens with tab
    b"% foo\nsc ed",     # separating tokens with space
    b"% foo\nsc\ned",    # separating tokens with newline
    b"% foo\nsc\red",    # separating tokens with linefeed
    b"% foo",            # glyph name only
    b"% foo bar",        # extra data after glyph name
])
def test_autohint_good_glyph(glyph):
    result = _psautohint.autohint(INFO, glyph)
    assert result == b"% foo\nsc\ned\n"


@pytest.mark.parametrize("glyph", [
    b"% foo\ncf",            # unknown operator
    b"% foo\n" + 80 * b"f",  # too long unknown operator
    b"% " + 65 * b"A",       # too long glyph name
    b"% foo\n10 ",           # number left on stack at end of glyph
    b"% foo\n0a 0 rm\ned",   # bad number terminator
    b"% foo\nry",            # stack underflow
    b"% foo\n$",             # unexpected character
])
def test_autohint_bad_glyph(glyph):
    with pytest.raises(_psautohint.error):
        _psautohint.autohint(INFO, glyph)


@pytest.mark.parametrize("glyphs", [
    (b"cf", b"cf"),
])
def test_autohintmm_bad_glyphs(glyphs):
    with pytest.raises(_psautohint.error):
        _psautohint.autohintmm(glyphs, (NAME, NAME))


@pytest.mark.parametrize("info", [
    b"HCounterChars [" + b" ".join(b"A" * i for i in range(16)) + b"]",
    b"VCounterChars [" + b" ".join(b"A" * i for i in range(16)) + b"]",
])
def test_autohint_too_many_counter_glyphs(info):
    _psautohint.autohint(info, GLYPH)



# Copyright 2014 Adobe. All rights reserved.

"""
Auto-hinting program for PostScript, OpenType/CFF and UFO fonts.
"""

import argparse
import logging
import os
import re
import subprocess
import sys
import textwrap

from . import __version__, get_font_format
from .autohint import ACOptions, hintFiles


FONTINFO_FILE_NAME = 'fontinfo'

GENERAL_INFO = """
psautohint is a tool for hinting fonts. It supports the following formats:
Type 1 (PFA and PFB), CFF (a.k.a. Type 2), OTF (name- and CID-keyed), and UFO.

By default, psautohint will hint all glyphs in the input font. Options are
available to specify a subset of the glyphs for hinting and other settings.

Note that the hinting results are better if the font's alignment zones are set
properly. For a font supporting a bicameral script (e.g. Latin, Greek, or
Cyrillic) the zones should capture the baseline, the x-height and the capital
height, and respective overshoots of the glyphs, at the very least.
Additionally, ascender, descender, small cap, and figures heights may also be
considered.

The reports provided by the stemHist tool can be useful for choosing alignment
zone and stem width values.
"""

FDDICT_DOC = r"""
By default, psautohint uses the font's global alignment zones and stem widths
when hinting each glyph. However, if there is a file named 'fontinfo' in the
same directory as the input font file, psautohint will search it for
definitions of sets of alignment zones (a.k.a 'FDDict'), and for matching
lists of glyphs to which each FDDict should be applied to. This approach
allows a group of glyphs to be hinted using a different set of zones and stem
widths than other glyphs. This solution isn't as robust as having multiple
hint dictionaries --supported in CID fonts-- in the font, as the final
name-keyed font can only have one set of alignment zones, but it does allow
for improved hinting when different sets of glyphs need different alignment
zones.

If FDDict definitions are used, then the global alignment zones and stem widths
in the source font will be ignored. For any glyphs not covered by an explicit
FDDict definition, psautohint will synthesize a dummy FDDict, where the zones
are set outside of the font's bounding box, so not to influence the hinting.
This is desirable for glyphs that have no features that need to be aligned.

If psautohint finds an FDDict named 'FinalFont', it will write that set of
values to the output font. Otherwise, it will merge all the alignment zones
and stem widths as the union of all FDDict definitions. If this merging fails
because some of the alignment zones and stem widths overlap, then the font
developer will have to provide a 'FinalFont' FDDict that explicitly defines
which stems and zones to use in the hinted output font.

To use a FDDict, one must define both the values of alignment zones and stem
widths, and the set of glyphs to apply it to. The FDDict must be defined in
the file before the set of glyphs which belong to it. Both the FDDict and the
glyph set define a name; an FDDict is applied to the glyph set with the same
name.

Running psautohint with the option --print-dflt-fddict will provide the list
of default FDDict values for the source font. These can be used as a starting
point for alternate FDDict definitions.

Running psautohint with the option --print-list-fddict will provide all the
user-defined FDDict defintions, as well as the list of glyphs associated with
each of them. One can use these to confirm the values, and to check which
glyphs are assigned to which FDDict. In particular, one should check the glyph
list of the FDDict named 'No Alignment Zones'; this list contains the glyphs
that did not match any of the search terms of the user-defined FDDicts.

The definitions use the following syntax:

begin FDDict <name>
    <key-1> <value-1>
    <key-2> <value-2>
    ...
    <key-n> <value-n>
end FDDict <name>

begin GlyphSet <name>
    <glyphname-1> <glyphname-2> ...
    <glyphname-n>
end GlyphSet <name>

The glyph names may be either a real glyph name, or a regular expression
designed to match several names.

An abbreviated regex primer:
^ ..... Matches at the start of the glyph name
$ ..... Matches at the end
[aABb]  Matches any one character in the set a, A, b, B
[A-Z]   Matches any one character in the set comprising the range from A-Z
[abA-Z] Matches any one character in the set a, b, and the characters
          in the range from A-Z
. ..... Matches any single character
+ ..... Maches whatever preceded it one or more times
* ..... Matches whatever preceded it none or more times.
\ ..... An escape character that includes the following character without
          the second one being interpreted as a regex special character

Examples:
^[A-Z]$    Matches names with one character in the range from A-Z.
^[A-Z].+   Matches any name where the first character is in the range A-Z,
             and it is followed by one or more characters.
[A-Z].+    Matches any name with a character that is in the range A-Z and
             which is followed by one or more characters.
^[A-Z].*   Matches any name with one or more characters, and
             the first character is in the range A-Z
^.+\.smallcaps   Matches any name that contains ".smallcaps"
^.+\.smallcaps$  Matches any name that ends with ".smallcaps"
^.+\.s[0-24]0$   Matches any name that ends with ".s00",".s10",".s20" or ".s40"


Example FDDict and GlyphSet definitions
=======================================

begin FDDict ST_Smallcaps
    # It's good practice to put the non-hint stuff first
    OrigEmSqUnits 1000
    FlexOK true
    # This gets used as the hint dict name if the font
    # is eventually built as a CID font.
    FontName MyFont-Bold

    # Alignment zones
    # The first is a bottom zone, the rest are top zones
    BaselineOvershoot -20
    BaselineYCoord 0
    CapHeight 900
    CapOvershoot 20
    LcHeight 700
    LcOvershoot 15

    # Stem widths
    DominantV [236 267]
    DominantH [141 152]
end FDDict ST_Smallcaps


begin FDDict LM_Smallcaps
    OrigEmSqUnits 1000
    FlexOK true
    FontName MyFont-Bold
    BaselineOvershoot -25
    BaselineYCoord 0
    CapHeight 950
    CapOvershoot 25
    LcHeight 750
    LcOvershoot 21
    DominantV [236 267]
    DominantH [141 152]
end FDDict LM_Smallcaps


begin GlyphSet LM_Smallcaps
    [Ll]\S+\.smallcap  [Mm]\S+\.smallcap
end GlyphSet LM_Smallcaps


begin GlyphSet ST_Smallcaps
    [Tt]\S+\.smallcap  [Ss]\S+\.smallcap
end GlyphSet ST_Smallcaps

***************************************

Note that whitespace must exist between keywords and values, but is otherwise
ignored. '#' marks the start of a comment; all the text after this character
on a line is ignored. GlyphSet and FDDict definitions may be intermixed,
as long as any FDDict is defined before the GlyphSet which refers to it.

At least two BlueValue pairs (the 'BaselineYCoord' bottom zone and any top
zone), and DominantH and DominantV values must be provided. All other keywords
are optional.

The full set of recognized FDDict keywords is:

BlueValue pairs:
    # BaselineOvershoot is a bottom zone, the rest are top zones.
    BaselineYCoord
    BaselineOvershoot

    CapHeight
    CapOvershoot

    LcHeight
    LcOvershoot

    AscenderHeight
    AscenderOvershoot

    FigHeight
    FigOvershoot

    Height5
    Height5Overshoot

    Height6
    Height6Overshoot

OtherBlues pairs:
    Baseline5Overshoot
    Baseline5

    Baseline6Overshoot
    Baseline6

    SuperiorOvershoot
    SuperiorBaseline

    OrdinalOvershoot
    OrdinalBaseline

    DescenderOvershoot
    DescenderHeight

For zones which capture the bottom of a feature in a glyph --BaselineYCoord
and all OtherBlues values-- the value specifies the top of the zone, and the
'Overshoot' is a negative value which specifes the offset to the bottom of
the zone, e.g.
    BaselineYCoord 0
    BaselineOvershoot -12

For zones which capture the top of a feature in a glyph --all BlueValue values
except BaselineYCoord-- the value specifies the bottom of the zone, and the
'Overshoot' is a positive value which specifes the offset to the top of the
zone, e.g.
    Height6 800
    Height6Overshoot 20

Note also that there is no implied sequential order of values. Height6 may have
a value less than or equal to CapHeight.

The values for keywords in one FontDict definiton are completely independent
from the values used in another FontDict. There is no inheritance from one
definition to the next.

Miscellaneous values:
  FontName ..... PostScript font name.
                 Only used by makeotf when building a CID font.
  OrigEmSqUnits  Single value: size of em-square.
                 Only used by makeotf when building a CID font.
  LanguageGroup  0 or 1. Specifies whether counter hints for ideographic
                 glyphs should be applied.
                 Only used by makeotf when building a CID font.
  DominantV .... List of dominant vertical stems, in the form
                 [<stem-value-1> <stem-value-2> ...]
  DominantH .... List of dominant horizontal stems, in the form
                 [<stem-value-1> <stem-value-2> ...]
  FlexOK ....... true or false.
  VCounterChars  List of glyphs to which counter hints may be applied,
                 in the form [<glyph-name-1> <glyph-name-2> ...]
  HCounterChars  List of glyphs to which counter hints may be applied,
                 in the form [<glyph-name-1> <glyph-name-2> ...]

Counter hints help to keep the space between stems open and equal in size.
psautohint will try and add counter hints to only a short hard-coded list
of glyphs,
    V-counters: m, M, T, ellipsis
    H-counters: element, equivalence, notelement, divide

To extend this list, use the VCounterChars and HCounterChars keywords.
A maximum of 64 glyph names can be added to either the vertical or
horizontal list.

---
Note for cognoscenti: the psautohint program code ignores StdHW and StdVW
entries if DominantV and DominantH entries are present, so it omits writing the
Std[HV]W keywords to fontinfo file. Also, psautohint will add any non-duplicate
stem width values for StemSnap[HV] to the Dominant[HV] stem width list, but the
StemSnap[HV] entries are not necessary if the full list of stem widths are
supplied as values for the Dominant[HV] keywords. Hence it also writes the full
stem list for the Dominant[HV] keywords, and does not write the StemSnap[HV]
keywords, to the fontinfo file. This is technically not right, as Dominant[HV]
array is supposed to hold only two values, but the psautohint program is not
affected, and it can write fewer entries this way.
"""


def _read_txt_file(file_path):
    with open(file_path, encoding='utf-8') as f:
        return f.read()


def _expand_cid_name(glyph_name, name_aliases):
    glyphRange = glyph_name.split("-")
    if len(glyphRange) > 1:
        g1 = _expand_cid_name(glyphRange[0], name_aliases)
        g2 = _expand_cid_name(glyphRange[1], name_aliases)
        glyph_name = "%s-%s" % (g1, g2)

    elif glyph_name[0] == "/":
        glyph_name = "cid" + glyph_name[1:].zfill(5)
        if glyph_name == "cid00000":
            glyph_name = ".notdef"
            name_aliases[glyph_name] = "cid00000"

    elif glyph_name.startswith("cid") and (len(glyph_name) < 8):
        glyph_name = "cid" + glyph_name[3:].zfill(5)
        if glyph_name == "cid00000":
            glyph_name = ".notdef"
            name_aliases[glyph_name] = "cid00000"

    return glyph_name


def _process_glyph_list_arg(glyph_list, name_aliases):
    return [_expand_cid_name(n, name_aliases) for n in glyph_list]


class HintOptions(ACOptions):
    def __init__(self, pargs):
        super(HintOptions, self).__init__()
        self.inputPaths = pargs.font_paths
        self.outputPaths = pargs.output_paths
        self.reference_font = pargs.reference_font
        self.hintAll = pargs.hint_all_ufo
        self.allowChanges = pargs.allow_changes
        self.noFlex = pargs.no_flex
        self.noHintSub = pargs.no_hint_sub
        self.allow_no_blues = pargs.no_zones_stems
        self.logOnly = pargs.report_only
        self.printDefaultFDDict = pargs.print_dflt_fddict
        self.printFDDictList = pargs.print_list_fddict
        self.round_coords = not pargs.decimal
        self.writeToDefaultLayer = pargs.write_to_default_layer


class _CustomHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """
    Adds extra line between options
    """
    @staticmethod
    def __add_whitespace(i, i_wtsp, arg):
        if i == 0:
            return arg
        return (" " * i_wtsp) + arg

    def _split_lines(self, arg, width):
        arg_rows = arg.splitlines()
        for i, line in enumerate(arg_rows):
            search = re.search(r'\s*[0-9\-]{0,}\.?\s*', line)
            if line.strip() == "":
                arg_rows[i] = " "
            elif search:
                line_wtsp = search.end()
                lines = [self.__add_whitespace(j, line_wtsp, x)
                         for j, x in enumerate(textwrap.wrap(line, width))]
                arg_rows[i] = lines

        # [''] adds the extra line between args
        return [item for sublist in arg_rows for item in sublist] + ['']


class _AdditionalHelpAction(argparse.Action):
    """
    Based on argparse's _HelpAction and _VersionAction.
    """

    def __init__(self,
                 option_strings,
                 addl_help=None,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help=None):
        super(_AdditionalHelpAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self.addl_help = addl_help

    def __call__(self, parser, namespace, values, option_string=None):
        formatter = parser._get_formatter()
        formatter.add_text(self.addl_help)
        parser.exit(message=formatter.format_help())


class DuplicateMessageFilter(logging.Filter):
    """
    Suppresses any log message that was reported before in the same module and
    for the same logging level. We check for module and level number in
    addition to the message just in case, though checking the message only is
    probably enough.
    """

    def __init__(self):
        super(DuplicateMessageFilter, self).__init__()
        self.logs = set()

    def filter(self, record):
        current = (record.module, record.levelno, record.getMessage())
        if current in self.logs:
            return False
        self.logs.add(current)
        return True


def _split_comma_sequence(comma_str):
    return [item.strip() for item in comma_str.split(',')]


def _check_save_path(path_str):
    test_path = os.path.abspath(os.path.realpath(path_str))
    del_test_file = True
    check_path = test_path
    if os.path.isdir(test_path):
        check_path = os.path.join(test_path, "dummy")
    try:
        if os.path.exists(check_path):
            del_test_file = False
        open(check_path, 'a').close()
        if del_test_file:
            os.remove(check_path)
    except (IOError, OSError):
        raise argparse.ArgumentTypeError(
            f"{test_path} is not a valid path to write to.")
    return test_path


def _check_tx():
    try:
        subprocess.check_output(["tx", "-h"], stderr=subprocess.STDOUT)
        return True
    except (subprocess.CalledProcessError, OSError):
        return False


def _validate_font_paths(path_lst, parser):
    """
    Checks that all input paths are fonts, and that all are the same format.
    Returns the font format string.
    """
    has_tx = None
    format_set = set()
    for path in path_lst:
        font_format = get_font_format(path)
        base_name = os.path.basename(path)
        if font_format is None:
            parser.error(f"{base_name} is not a supported font format")
        if font_format in ("PFA", "PFB", "PFC"):
            if has_tx is None:
                has_tx = _check_tx()
            if not has_tx:
                parser.error(f"{base_name} font format requires 'tx'. "
                             "Please install 'afdko'.")
        format_set.add(font_format)
    if len(format_set) > 1:
        parser.error("the input fonts must be all of the same format")
    # get the set's single-item via tuple unpacking
    ft_format_str, = format_set
    return ft_format_str


def _validate_path(path_str):
    valid_path = os.path.abspath(os.path.realpath(path_str))
    if not os.path.exists(valid_path):
        raise argparse.ArgumentTypeError(
            f"{path_str} is not a valid path.")
    return valid_path


RE_COUNTER_CHARS = re.compile(
    r"([VH])CounterChars\s+[\(\[]\s*([^\)\r\n]+)[\)\]]")
RE_COMMENT_LINE = re.compile(r"#[^\r\n]+")


def _parse_fontinfo_file(options, fontinfo_path):
    """
    Collects VCounterChars and HCounterChars values from a fontinfo file,
    and updates the ACOptions class.
    """
    data = _read_txt_file(fontinfo_path)
    data = re.sub(RE_COMMENT_LINE, "", data)
    counterGlyphLists = RE_COUNTER_CHARS.findall(data)
    for entry in counterGlyphLists:
        glyphList = entry[1].split()
        if glyphList:
            if entry[0] == "V":
                options.vCounterGlyphs.extend(glyphList)
            else:
                options.hCounterGlyphs.extend(glyphList)


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=_CustomHelpFormatter,
        description=__doc__
    )
    parser_fonts = parser.add_argument(
        'font_paths',
        metavar='FONT',
        nargs='*',
        type=_validate_path,
        help='Type1/CFF/OTF/UFO font files'
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for extra-verbose mode.'
    )
    parser.add_argument(
        '-o',
        '--output',
        metavar='PATH',
        nargs='+',
        dest='output_paths',
        type=_check_save_path,
        help='specify a file to write the hints to\n'
             'The hints are written to a new font instead of modifying the '
             'original one.'
    )
    parser.add_argument(
        '-r',
        '--reference-font',
        metavar='PATH',
        type=_validate_path,
        help='reference font\n'
             'Font to be used as reference, when hinting multiple fonts '
             'compatibily.'
    )
    parser.add_argument(
        '-a',
        '--all',
        action='store_true',
        dest='hint_all_ufo',
        help='hint all glyphs\n'
             'All glyphs get hinted even if:\n'
             '- they have been hinted before, or\n'
             '- the outlines in the default layer have not changed\n'
             'This is a UFO-only option.'
    )
    parser.add_argument(
        '-w',
        '--write-to-default-layer',
        action='store_true',
        help='write hints to default layer. This is a UFO-only option.'
    )
    glyphs_parser = parser.add_mutually_exclusive_group()
    glyphs_parser.add_argument(
        '-g',
        '--glyphs',
        metavar='GLYPH_LIST',
        dest='glyphs_to_hint',
        type=_split_comma_sequence,
        default=[],
        help='comma-separated sequence of glyphs to hint\n'
             'The glyph identifiers may be glyph indexes, glyph names, or '
             'glyph CIDs. CID values must be prefixed with a forward slash.\n'
             'Examples:\n'
             '    psautohint -g A,B,C,69 MyFont.ufo\n'
             '    psautohint -g /103,/434,68 MyCIDFont'
    )
    glyphs_parser.add_argument(
        '--glyphs-file',
        metavar='PATH',
        dest='glyphs_to_hint_file',
        type=_validate_path,
        help='file containing a list of glyphs to hint\n'
             'The file must contain a comma-separated sequence of glyph '
             'identifiers.'
    )
    glyphs_parser.add_argument(
        '-x',
        '--exclude-glyphs',
        metavar='GLYPH_LIST',
        dest='glyphs_to_not_hint',
        type=_split_comma_sequence,
        default=[],
        help='comma-separated sequence of glyphs to NOT hint\n'
             "Counterpart to '--glyphs' option."
    )
    glyphs_parser.add_argument(
        '--exclude-glyphs-file',
        metavar='PATH',
        dest='glyphs_to_not_hint_file',
        type=_validate_path,
        help='file containing a list of glyphs to NOT hint\n'
             "Counterpart to '--glyphs-file' option."
    )
    report_parser = parser.add_mutually_exclusive_group()
    report_parser.add_argument(
        '-c',
        '--allow-changes',
        action='store_true',
        help='allow changes to the glyph outlines\n'
             'Paths are reordered to reduce hint substitution, and nearly '
             'straight curves are flattened.'
    )
    report_parser.add_argument(
        '--report-only',
        action='store_true',
        help='process the font without modifying it'
    )
    parser.add_argument(
        '--log',
        metavar='PATH',
        dest='log_path',
        type=_check_save_path,
        help='write output messages to a file'
    )
    parser.add_argument(
        '-d',
        '--decimal',
        action='store_true',
        help='use decimal coordinates'
    )
    parser.add_argument(
        '--no-flex',
        action='store_true',
        help='suppress generation of flex commands'
    )
    parser.add_argument(
        '--no-hint-sub',
        action='store_true',
        help='suppress hint substitution'
    )
    parser.add_argument(
        '--no-zones-stems',
        action='store_true',
        help='allow the font to have no alignment zones nor stem widths'
    )
    parser.add_argument(
        '--fontinfo-file',
        metavar='PATH',
        type=_validate_path,
        help='file containing hinting parameters\n'
             f"Default: '{FONTINFO_FILE_NAME}'"
    )
    parser.add_argument(
        '--print-dflt-fddict',
        action='store_true',
        help="print the font's default FDDict set\n"
             'This can be used as a starting point for building FDDict '
             "sets. Use '--doc-fddict' option for more info."
    )
    parser.add_argument(
        '--print-list-fddict',
        action='store_true',
        help='print the list of user-defined FDDict sets, and the glyphs '
             'associated with each of them\n'
             'This is useful for checking the FDDict sets and their glyph-'
             "matching definitions. Use '--doc-fddict' option for more info."
    )
    parser.add_argument(
        '--doc-fddict',
        action=_AdditionalHelpAction,
        help="show a description of the format for defining sets of "
             "alternate alignment zones in a 'fontinfo' file and exit",
        addl_help=FDDICT_DOC
    )
    parser.add_argument(
        '--info',
        action=_AdditionalHelpAction,
        help="show program's general info and exit",
        addl_help=GENERAL_INFO
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '--traceback',
        action='store_true',
        help="show traceback for exceptions.",
    )
    parser.add_argument(
        '--test',
        action='store_true',
        help="toggle settings needed for testing",
    )

    parsed_args = parser.parse_args(args)

    if parsed_args.verbose == 0:
        log_level = logging.WARNING
    elif parsed_args.verbose == 1:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG

    logging.basicConfig(format="%(levelname)s: %(message)s", level=log_level,
                        filename=parsed_args.log_path)

    if parsed_args.test is False:
        # Filter duplicate logging messages only when not running the tests
        for handler in logging.root.handlers:
            handler.addFilter(DuplicateMessageFilter())

    if (not len(parsed_args.font_paths or []) and
            len(parsed_args.output_paths or [])):
        # allow "psautohint -o outputpath inputpath"
        # see https://github.com/adobe-type-tools/psautohint/issues/129
        half = len(parsed_args.output_paths) // 2
        parsed_args.font_paths = [
            parsed_args.output_paths.pop(half) for _ in range(half)]

    if not len(parsed_args.font_paths):
        parser.error(
            f"the following arguments are required: {parser_fonts.metavar}")

    if (parsed_args.output_paths and
            len(parsed_args.font_paths) != len(parsed_args.output_paths)):
        parser.error("number of input and output fonts differ")

    if parsed_args.reference_font in parsed_args.font_paths:
        parser.error("the reference font cannot also be a font input")

    if len(parsed_args.font_paths) != len(set(parsed_args.font_paths)):
        parser.error("found duplicate font inputs")

    if parsed_args.reference_font:
        all_font_paths = parsed_args.font_paths + [parsed_args.reference_font]
    else:
        all_font_paths = parsed_args.font_paths

    options = HintOptions(parsed_args)

    options.font_format = _validate_font_paths(all_font_paths, parser)

    if parsed_args.glyphs_to_hint:
        options.glyphList = _process_glyph_list_arg(
            parsed_args.glyphs_to_hint, options.nameAliases)
    elif parsed_args.glyphs_to_not_hint:
        options.excludeGlyphList = True
        options.glyphList = _process_glyph_list_arg(
            parsed_args.glyphs_to_not_hint, options.nameAliases)
    elif parsed_args.glyphs_to_hint_file:
        options.glyphList = _process_glyph_list_arg(
            _read_txt_file(parsed_args.glyphs_to_hint_file),
            options.nameAliases)
    elif parsed_args.glyphs_to_not_hint_file:
        options.excludeGlyphList = True
        options.glyphList = _process_glyph_list_arg(
            _read_txt_file(parsed_args.glyphs_to_not_hint_file),
            options.nameAliases)

    if not parsed_args.fontinfo_file:
        fontinfo_path = os.path.join(os.path.dirname(all_font_paths[0]),
                                     FONTINFO_FILE_NAME)
        if os.path.isfile(fontinfo_path):
            _parse_fontinfo_file(options, fontinfo_path)
    else:
        _parse_fontinfo_file(options, parsed_args.fontinfo_file)

    return options, parsed_args


def main(args=None):
    options, pargs = get_options(args)

    try:
        hintFiles(options)
    except Exception as ex:
        if pargs.traceback:
            raise
        logging.error(ex)
        return 1


class ReportOptions(ACOptions):
    def __init__(self, pargs):
        super(ReportOptions, self).__init__()
        self.hintAll = True
        self.noFlex = True
        self.allow_no_blues = True
        self.logOnly = True
        self.inputPaths = pargs.font_paths
        self.outputPaths = pargs.output_paths
        self.report_stems = not pargs.alignment_zones
        self.report_zones = pargs.alignment_zones
        self.report_all_stems = pargs.all_stems


def get_stemhist_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=_CustomHelpFormatter,
        description='Stem and Alignment zones report for PostScript, '
                    'OpenType/CFF and UFO fonts.'
    )
    parser.add_argument(
        'font_paths',
        metavar='FONT',
        nargs='+',
        type=_validate_path,
        help='Type1/CFF/OTF/UFO font files'
    )
    parser.add_argument(
        '-o',
        '--output',
        metavar='PATH',
        nargs='+',
        dest='output_paths',
        type=_check_save_path,
        help='When this is specified, the path argument is used as the base '
             'path for the reports. Otherwise, the font file path is used as '
             'the base path. Several reports are written, using the base path '
             'name plus an extension.\n'
             'Without the -z option, the report files are:\n'
             '   <base path>.hstm.txt  The horizontal stems.\n'
             '   <base path>.vstm.txt  The vertical stems.\n'
             'With the -z option, the report files are:\n'
             '   <base path>.top.txt   The top zones.\n'
             '   <base path>.bot.txt   The bottom zones.'
    )
    parser.add_argument(
        '-z',
        '--alignment-zones',
        action='store_true',
        dest='alignment_zones',
        help='Print alignment zone report rather than stem report.'
    )
    parser.add_argument(
        '-a',
        '--all',
        action='store_true',
        dest='all_stems',
        help='Include stems formed by curved line segments; by default, '
             'includes only stems formed by straight line segments.'
    )
    glyphs_parser = parser.add_mutually_exclusive_group()
    glyphs_parser.add_argument(
        '-g',
        '--glyphs',
        metavar='GLYPH_LIST',
        dest='glyphs_to_hint',
        type=_split_comma_sequence,
        default=[],
        help='comma-separated sequence of glyphs to process\n'
             'The glyph identifiers may be glyph indexes, glyph names, or '
             'glyph CIDs. CID values must be prefixed with a forward slash.\n'
             'Examples:\n'
             '    psautohint -g A,B,C,69 MyFont.ufo\n'
             '    psautohint -g /103,/434,68 MyCIDFont'
    )
    glyphs_parser.add_argument(
        '--glyphs-file',
        metavar='PATH',
        dest='glyphs_to_hint_file',
        type=_validate_path,
        help='file containing a list of glyphs to process\n'
             'The file must contain a comma-separated sequence of glyph '
             'identifiers.'
    )
    glyphs_parser.add_argument(
        '-x',
        '--exclude-glyphs',
        metavar='GLYPH_LIST',
        dest='glyphs_to_not_hint',
        type=_split_comma_sequence,
        default=[],
        help='comma-separated sequence of glyphs to NOT process\n'
             "Counterpart to '--glyphs' option."
    )
    glyphs_parser.add_argument(
        '--exclude-glyphs-file',
        metavar='PATH',
        dest='glyphs_to_not_hint_file',
        type=_validate_path,
        help='file containing a list of glyphs to NOT process\n'
             "Counterpart to '--glyphs-file' option."
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for extra-verbose mode.'
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '--traceback',
        action='store_true',
        help="show traceback for exceptions.",
    )
    parsed_args = parser.parse_args(args)

    if parsed_args.verbose == 0:
        log_level = logging.WARNING
    elif parsed_args.verbose == 1:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG

    logging.basicConfig(format="%(levelname)s: %(message)s", level=log_level)
    for handler in logging.root.handlers:
        handler.addFilter(DuplicateMessageFilter())

    if (parsed_args.output_paths and
            len(parsed_args.font_paths) != len(parsed_args.output_paths)):
        parser.error("number of input and output fonts differ")

    if len(parsed_args.font_paths) != len(set(parsed_args.font_paths)):
        parser.error("found duplicate font inputs")

    options = ReportOptions(parsed_args)

    options.font_format = _validate_font_paths(parsed_args.font_paths, parser)

    if parsed_args.glyphs_to_hint:
        options.glyphList = _process_glyph_list_arg(
            parsed_args.glyphs_to_hint, options.nameAliases)
    elif parsed_args.glyphs_to_not_hint:
        options.excludeGlyphList = True
        options.glyphList = _process_glyph_list_arg(
            parsed_args.glyphs_to_not_hint, options.nameAliases)
    elif parsed_args.glyphs_to_hint_file:
        options.glyphList = _process_glyph_list_arg(
            _read_txt_file(parsed_args.glyphs_to_hint_file),
            options.nameAliases)
    elif parsed_args.glyphs_to_not_hint_file:
        options.excludeGlyphList = True
        options.glyphList = _process_glyph_list_arg(
            _read_txt_file(parsed_args.glyphs_to_not_hint_file),
            options.nameAliases)

    return options, parsed_args


def stemhist(args=None):
    options, pargs = get_stemhist_options(args)

    try:
        hintFiles(options)
    except Exception as ex:
        if pargs.traceback:
            raise
        logging.error(ex)
        return 1


if __name__ == '__main__':
    sys.exit(main())



import glob
from os.path import basename
import pytest

from fontTools.misc.xmlWriter import XMLWriter
from fontTools.ttLib import TTFont
from psautohint.autohint import ACOptions, ACHintError, hintFiles

from .differ import main as differ
from . import make_temp_copy, DATA_DIR


class Options(ACOptions):

    def __init__(self, reference, inpaths, outpaths):
        super(Options, self).__init__()
        self.inputPaths = inpaths
        self.outputPaths = outpaths
        self.reference_font = reference
        self.hintAll = True
        self.verbose = False
        self.verbose = False


@pytest.mark.parametrize("base", glob.glob("%s/*/Masters" % DATA_DIR))
def test_mmufo(base, tmpdir):
    paths = sorted(glob.glob(base + "/*.ufo"))
    # the reference font is modified in-place, make a temp copy first
    reference = make_temp_copy(tmpdir, paths[0])
    inpaths = paths[1:]
    outpaths = [str(tmpdir / basename(p)) + ".out" for p in inpaths]

    options = Options(reference, inpaths, outpaths)
    hintFiles(options)

    for inpath, outpath in zip(inpaths, outpaths):
        assert differ([inpath, outpath])


@pytest.mark.parametrize("base", glob.glob("%s/*/OTFMasters" % DATA_DIR))
def test_mmotf(base, tmpdir):
    paths = sorted(glob.glob(base + "/*.otf"))
    # the reference font is modified in-place, make a temp copy first
    reference = make_temp_copy(tmpdir, paths[0])
    inpaths = paths[1:]
    outpaths = [str(tmpdir / basename(p)) + ".out" for p in inpaths]

    options = Options(reference, inpaths, outpaths)
    hintFiles(options)

    refs = [p + ".ref" for p in paths]
    for ref, out in zip(refs, [reference] + outpaths):
        for path in (ref, out):
            font = TTFont(path)
            assert "CFF " in font
            writer = XMLWriter(str(tmpdir / basename(path)) + ".xml")
            font["CFF "].toXML(writer, font)
            writer.close()

        assert differ([str(tmpdir / basename(ref)) + ".xml",
                       str(tmpdir / basename(out)) + ".xml"])


@pytest.mark.parametrize("otf", glob.glob("%s/vf_tests/*.otf" % DATA_DIR))
def test_vfotf(otf, tmpdir):
    out = str(tmpdir / basename(otf)) + ".out"
    options = Options(None, [otf], [out])
    options.allow_no_blues = True
    hintFiles(options)

    for path in (otf, out):
        font = TTFont(path)
        assert "CFF2" in font
        writer = XMLWriter(str(tmpdir / basename(path)) + ".xml")
        font["CFF2"].toXML(writer, font)
        writer.close()
    assert differ([str(tmpdir / basename(otf)) + ".xml",
                   str(tmpdir / basename(out)) + ".xml"])


def test_incompatible_masters(tmpdir):
    base = "%s/source-serif-pro/" % DATA_DIR
    paths = [base + "Light/font.ufo", base + "Black/font.ufo"]
    # the reference font is modified in-place, make a temp copy first
    reference = make_temp_copy(tmpdir, paths[0])
    inpaths = paths[1:]
    outpaths = [str(tmpdir / p) for p in inpaths]

    options = Options(reference, inpaths, outpaths)
    with pytest.raises(ACHintError):
        hintFiles(options)


def test_sparse_mmotf(tmpdir):
    base = "%s/sparse_masters" % DATA_DIR
    paths = sorted(glob.glob(base + "/*.otf"))
    # the reference font is modified in-place, make a temp copy first
    # MasterSet_Kanji-w0.00.otf has to be the reference font.
    reference = make_temp_copy(tmpdir, paths[0])
    inpaths = paths[1:]
    outpaths = [str(tmpdir / basename(p)) + ".out" for p in inpaths]

    options = Options(reference, inpaths, outpaths)
    options.allow_no_blues = True
    hintFiles(options)

    refs = [p + ".ref" for p in paths]
    for ref, out in zip(refs, [reference] + outpaths):
        for path in (ref, out):
            font = TTFont(path)
            assert "CFF " in font
            writer = XMLWriter(str(tmpdir / basename(path)) + ".xml")
            font["CFF "].toXML(writer, font)
            writer.close()

        assert differ([str(tmpdir / basename(ref)) + ".xml",
                       str(tmpdir / basename(out)) + ".xml"])



import glob
import os
import pytest

from psautohint.autohint import ACOptions, openFile
from psautohint import hint_bez_glyph

from . import DATA_DIR


class BezFontData:
    def __init__(self, path):
        self._path = path
        self._info = None
        self._glyphs = {}

    class FontInfo:
        def __init__(self, info):
            self._info = info

        def getFontInfo(self):
            return self._info

    def getFontInfo(self, allow_no_blues, noFlex, vCounterGlyphs,
                    hCounterGlyphs):
        if self._info is None:
            with open(os.path.join(self._path, "fontinfo")) as fp:
                self._info = self.FontInfo(fp.read())
        return self._info

    def convertToBez(self, glyphName, read_hints, doAll=False):
        if glyphName not in self._glyphs:
            with open(os.path.join(self._path, glyphName + ".bez")) as fp:
                self._glyphs[glyphName] = fp.read()
        return self._glyphs[glyphName]

    def getGlyphList(self):
        files = glob.glob(self._path + "/*.bez")
        files += glob.glob(self._path + "/.*.bez")
        return [os.path.splitext(os.path.basename(f))[0] for f in files]


def open_font(path):
    if not path.endswith(".bez"):
        font = openFile(path, ACOptions())
    else:
        font = BezFontData(path)

    return font


def get_font_info(font, path):
    info = font.getFontInfo(False, False, [], [])
    # Sort to normalize the order.
    info = sorted(info.getFontInfo().split("\n"))
    return "\n".join(info)


def normalize_glyph(glyph, name):
    """Dump Bez format normalizer."""

    # strip comments
    hhints = []
    vhints = []
    path = []
    for line in glyph.split("\n"):
        line = line.split("%")[0].strip()
        if not line:
            pass

        elif line.endswith("rb"):
            hhints.append(line)
        elif line.endswith("ry"):
            vhints.append(line)
        else:
            path.append(line)
    hhints.sort()
    vhints.sort()
    return "\n".join(["% " + name] + hhints + vhints + path)


def get_glyph(font, name):
    glyph = font.convertToBez(name, True, True)
    return normalize_glyph(glyph, name)


@pytest.mark.parametrize("unhinted,hinted", zip(
    glob.glob("%s/unhinted/*.[uo][ft][of]" % DATA_DIR),
    glob.glob("%s/hinted/*.[uo][ft][of]" % DATA_DIR),
))
def test_bez(unhinted, hinted):
    unhinted_base = os.path.splitext(unhinted)[0]
    hinted_base = os.path.splitext(hinted)[0]

    bez_font = open_font(unhinted_base + ".bez")
    otf_font = open_font(unhinted_base + ".otf")
    ufo_font = open_font(unhinted_base + ".ufo")
    hinted_bez_font = open_font(hinted_base + ".bez")
    hinted_otf_font = open_font(hinted_base + ".otf")

    bez_info = get_font_info(bez_font, unhinted_base + ".bez")
    otf_info = get_font_info(otf_font, unhinted_base + ".otf")
    ufo_info = get_font_info(ufo_font, unhinted_base + ".ufo")
    assert otf_info == bez_info == ufo_info

    names = sorted(otf_font.getGlyphList())
    assert all(sorted(f.getGlyphList()) == names for f in [bez_font,
                                                           otf_font,
                                                           ufo_font,
                                                           hinted_bez_font,
                                                           hinted_otf_font])
    for name in names:
        if name == ".notdef":
            continue
        bez_glyph = get_glyph(bez_font, name)
        otf_glyph = get_glyph(otf_font, name)
        ufo_glyph = get_glyph(ufo_font, name)
        assert otf_glyph == bez_glyph == ufo_glyph

        hinted_bez_glyph = get_glyph(hinted_bez_font, name)
        hinted_otf_glyph = get_glyph(hinted_otf_font, name)
        assert hinted_otf_glyph == hinted_bez_glyph

        result = hint_bez_glyph(bez_info, bez_glyph)
        assert normalize_glyph(result, name) == hinted_bez_glyph



from psautohint.ufoFont import UFOFontData

from . import DATA_DIR


def test_incomplete_glyphorder():
    path = "%s/dummy/incomplete_glyphorder.ufo" % DATA_DIR
    font = UFOFontData(path, False, True)

    assert len(font.getGlyphList()) == 95
    assert "ampersand" in font.getGlyphList()



# Copyright 2018 Adobe. All rights reserved.

"""
Helper script for diff'ing files.
Used as part of the integration tests.
"""

import argparse
import difflib
import filecmp
import logging
import os
import re
import sys

__version__ = '0.3.3'

logger = logging.getLogger('differ')


TXT_MODE = 'txt'
BIN_MODE = 'bin'
SPLIT_MARKER = '_+:+_'
DFLT_ENC = 'utf-8'


class Differ(object):

    def __init__(self, opts):
        self.mode = opts.mode
        self.path1 = opts.path1
        self.path2 = opts.path2
        # tuple of strings containing the starts of lines to skip
        self.skip_strings = opts.skip_strings
        # tuple of integers for the line numbers to skip
        self.skip_lines = opts.skip_lines
        # regex pattern matching the beginning of line
        self.skip_regex = opts.skip_regex
        self.encoding = opts.encoding

    def diff_paths(self):
        """
        Diffs the contents of two paths using the parameters provided.
        Returns True if the contents match, and False if they don't.
        """
        if os.path.isdir(self.path1):
            return self._diff_dirs()
        else:  # paths are files
            if self.mode == TXT_MODE:
                return self._diff_txt_files(self.path1, self.path2)
            elif self.mode == BIN_MODE:
                return filecmp.cmp(self.path1, self.path2)

    def _diff_txt_files(self, path1, path2):
        """
        Diffs two text-based files using difflib.
        Returns True if the contents match, and False if they don't.

        NOTE: This method CANNOT use self.path1 and self.path2 because it
        gets called from _diff_dirs().
        """
        expected = self._read_txt_file(path1)
        actual = self._read_txt_file(path2)
        if actual != expected:
            for line in difflib.unified_diff(expected, actual,
                                             fromfile=path1,
                                             tofile=path2, n=1):
                sys.stdout.write(line)
            return False
        return True

    def _read_txt_file(self, path):
        """
        Reads a text file and returns a list of its lines.
        """
        # Hard code a first line; this way the difflib results start
        # from 1 instead of zero, thus matching the file's line numbers
        lines = ['']
        try:
            with open(path, "r", encoding=self.encoding) as f:
                for i, line in enumerate(f.readlines(), 1):
                    # Skip lines that change, such as timestamps
                    if self._line_to_skip(line):
                        logger.debug(f"Matched begin of line. "
                                     f"Skipped: {line.rstrip()}")
                        # Blank the line instead of actually skipping (via
                        # 'continue'); this way the difflib results show the
                        # correct line numbers
                        line = ''
                    # Skip specific lines, referenced by number
                    elif i in self.skip_lines:
                        logger.debug(f"Matched line #{i}. "
                                     f"Skipped: {line.rstrip()}")
                        line = ''
                    # Skip lines that match regex
                    elif self.skip_regex and self.skip_regex.match(line):
                        logger.debug(f"Matched regex begin of line. "
                                     f"Skipped: {line.rstrip()}")
                        line = ''
                    # Use os-native line separator to enable running difflib
                    lines.append(line.rstrip() + os.linesep)
        except UnicodeDecodeError:
            logger.error(f"Couldn't read text file using '{self.encoding}' "
                         f"encoding.\n      File path: {path}")
            sys.exit(1)
        return lines

    def _line_to_skip(self, line):
        """
        Loops over the skip items.
        Returns True if the beginning of the line matches a skip item.
        """
        for item in self.skip_strings:
            if line.startswith(item):
                return True
        return False

    def _diff_dirs(self):
        """
        Diffs two folders containing files.
        Returns True if all files match. Returns False if the folders' contents
        don't match, or as soon as one non-matching file is found.
        """
        all_rel_file_paths = self._compare_dir_contents()
        if all_rel_file_paths is None:
            return False
        for rel_file_path in all_rel_file_paths:
            path1 = self.path1 + rel_file_path
            assert os.path.exists(path1), f"Not a valid path1: {path1}"
            path2 = self.path2 + rel_file_path
            assert os.path.exists(path2), f"Not a valid path2: {path2}"
            if self.mode == BIN_MODE:
                diff_result = filecmp.cmp(path1, path2)
            else:
                diff_result = self._diff_txt_files(path1, path2)
            if not diff_result:
                logger.debug(f"Non-matching file: {rel_file_path}")
                return False
        return True

    def _report_dir_diffs(self, all_paths1, all_paths2):
        """
        Returns a string listing the paths that exist in folder 1 but not in 2,
        and vice-versa.
        """
        diffs_str = ''
        set_1st = set(all_paths1)
        set_2nd = set(all_paths2)
        diff1 = sorted(set_1st - set_2nd)
        diff2 = sorted(set_2nd - set_1st)
        if diff1:
            dir1 = os.path.basename(self.path1)
            dj1 = '\n    '.join(diff1)
            diffs_str += (f"\n  In 1st folder ({dir1}) but not in 2nd:"
                          f"\n    {dj1}")
        if diff2:
            dir2 = os.path.basename(self.path2)
            dj2 = '\n    '.join(diff2)
            diffs_str += (f"\n  In 2nd folder ({dir2}) but not in 1st:"
                          f"\n    {dj2}")
        return diffs_str

    def _compare_dir_contents(self):
        """
        Checks if two directory trees have the same files and folders.
        Returns a list of relative paths to all files if the dirs' contents
        match, and None if they don't.
        """
        all_paths1 = self._get_all_file_paths_in_dir_tree(self.path1)
        all_paths2 = self._get_all_file_paths_in_dir_tree(self.path2)
        if all_paths1 != all_paths2:
            dd = self._report_dir_diffs(all_paths1, all_paths2)
            logger.info(f"Folders' contents don't match.{dd}")
            return None
        return all_paths1

    @staticmethod
    def _get_all_file_paths_in_dir_tree(start_path):
        """
        Returns a list of relative paths of all files in a directory tree.
        The list's items are ordered top-down according to the tree.
        """
        all_paths = []
        for dir_name, _, file_names in os.walk(start_path):
            all_paths.extend(
                [os.path.join(dir_name, f_name) for f_name in file_names])

        # Make the paths relative, and enforce order.
        all_paths = sorted(
            [path.replace(start_path, '') for path in all_paths])

        logger.debug(f"All paths: {all_paths}")

        return all_paths


def _get_path_kind(pth):
    """
    Returns a string describing the kind of path.
    Possible values are 'file', 'folder', and 'invalid'.
    """
    try:
        if os.path.isfile(pth):
            return 'file'
        elif os.path.isdir(pth):
            return 'folder'
        elif not os.path.exists(pth):
            return 'invalid'
    except TypeError:
        return 'invalid'


def _paths_are_same_kind(path1, path2):
    """
    Checks that both paths are either files or folders.
    Returns boolean.
    """
    if all([os.path.isfile(path) for path in (path1, path2)]):
        return True
    elif all([os.path.isdir(path) for path in (path1, path2)]):
        return True
    return False


def _validate_path(path_str):
    valid_path = os.path.abspath(os.path.realpath(path_str))
    if not os.path.exists(valid_path):
        raise argparse.ArgumentTypeError(
            f"{path_str} is not a valid path.")
    return valid_path


def _split_string_sequence(str_seq):
    return tuple(str_seq.split(SPLIT_MARKER))


def _split_num_range_or_delta(num_str):
    num_range = num_str.split('-') + ['-']
    num_delta = num_str.split('+') + ['+']
    if len(num_range) == 3:
        return num_range
    elif len(num_delta) == 3:
        return num_delta
    elif num_str.isnumeric():
        return num_str
    else:
        raise argparse.ArgumentTypeError(
            f"Invalid number range or delta: {num_str}")


def _convert_to_int(num_str):
    try:
        return int(num_str)
    except ValueError:
        raise argparse.ArgumentTypeError(f"Not a number: {num_str}")


def _expand_num_range_or_delta(num_str_lst):
    start_num = _convert_to_int(num_str_lst[0])
    rng_dlt_num = _convert_to_int(num_str_lst[1])
    sign = num_str_lst[2]
    if sign == '+':
        return list(range(start_num, start_num + rng_dlt_num + 1))
    else:  # sign == '-'
        if not (rng_dlt_num >= start_num):
            raise argparse.ArgumentTypeError(
                f"The start of range value is larger than the end of range "
                f"value: {start_num}-{rng_dlt_num}")
        return list(range(start_num, rng_dlt_num + 1))


def _convert_seq_to_ints(num_seq):
    seq = []
    for item in num_seq:
        if isinstance(item, list):
            seq.extend(_expand_num_range_or_delta(item))
        else:
            seq.append(_convert_to_int(item))
    return sorted(set(seq))


def _split_linenumber_sequence(str_seq):
    num_seq = [_split_num_range_or_delta(item) for item in str_seq.split(',')]
    return tuple(_convert_seq_to_ints(num_seq))


def _compile_regex(str_seq):
    if not str_seq.startswith('^'):
        raise argparse.ArgumentTypeError(
            "The expression must start with the caret '^' character")
    try:
        return re.compile(str_seq)
    except re.error as err:
        raise argparse.ArgumentTypeError(
            f'The expression is invalid: {err}')


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for debug mode'
    )
    parser.add_argument(
        '-m',
        '--mode',
        default=TXT_MODE,
        choices=(TXT_MODE, BIN_MODE),
        help='diff mode (default: %(default)s)'
    )
    parser.add_argument(
        '-s',
        '--string',
        dest='skip_strings',
        type=_split_string_sequence,
        default=(),
        help=f'string for matching the beginning of a line to skip\n'
             f'For multiple strings, separate them with {SPLIT_MARKER}'
    )
    parser.add_argument(
        '-l',
        '--line',
        dest='skip_lines',
        type=_split_linenumber_sequence,
        default=(),
        help='number of a line to skip\n'
             'For multiple line numbers, separate them with a comma (,).\n'
             'For ranges of lines, use a minus (-) between two numbers.\n'
             'For a line delta, use a plus (+) between two numbers.'
    )
    parser.add_argument(
        '-r',
        '--regex',
        dest='skip_regex',
        type=_compile_regex,
        help='regular expression matching the beginning of a line to skip\n'
             "The expression must start with the caret '^' character, "
             'and characters such as backslash, semicolon, and space need '
             'to be escaped.'
    )
    parser.add_argument(
        '-e',
        '--encoding',
        default=DFLT_ENC,
        choices=(DFLT_ENC, 'macroman'),
        help='encoding to use when opening text files (default: %(default)s)'
    )
    parser.add_argument(
        'path1',
        metavar='PATH1',
        type=_validate_path,
        help='1st path for comparison'
    )
    parser.add_argument(
        'path2',
        metavar='PATH2',
        type=_validate_path,
        help='2nd path for comparison'
    )
    options = parser.parse_args(args)

    if not options.verbose:
        level = "WARNING"
    elif options.verbose == 1:
        level = "INFO"
    else:
        level = "DEBUG"
    logging.basicConfig(level=level)

    if not _paths_are_same_kind(options.path1, options.path2):
        kp1 = _get_path_kind(options.path1)
        kp2 = _get_path_kind(options.path2)
        parser.error(f"The paths are not of the same kind. "
                     f"Path1's kind is {kp1}. "
                     f"Path2's kind is {kp2}.")

    logger.debug(f"Line numbers: {options.skip_lines}")
    regexpat = getattr(options.skip_regex, 'pattern', None)
    logger.debug(f"Regular expression: {regexpat}")

    return options


def main(args=None):
    """
    Returns True if the inputs match, and False if they don't.
    """
    opts = get_options(args)

    differ = Differ(opts)
    return differ.diff_paths()


if __name__ == "__main__":
    sys.exit(0 if main() else 1)



# Copyright 2014 Adobe. All rights reserved.

"""
Utilities for converting between T2 charstrings and the bez data format.
"""

import copy
import logging
import os
import re
import subprocess
import tempfile
import itertools

from fontTools.misc.psCharStrings import (T2OutlineExtractor,
                                          SimpleT2Decompiler)
from fontTools.ttLib import TTFont, newTable
from fontTools.misc.roundTools import noRound, otRound
from fontTools.varLib.varStore import VarStoreInstancer
from fontTools.varLib.cff import CFF2CharStringMergePen, MergeOutlineExtractor
# import subset.cff is needed to load the implementation for
# CFF.desubroutinize: the module adds this class method to the CFF and CFF2
# classes.
import fontTools.subset.cff

from . import fdTools, FontParseError

# keep linting tools quiet about unused import
assert fontTools.subset.cff is not None

log = logging.getLogger(__name__)

kStackLimit = 46
kStemLimit = 96


class SEACError(Exception):
    pass


def _add_method(*clazzes):
    """Returns a decorator function that adds a new method to one or
    more classes."""
    def wrapper(method):
        done = []
        for clazz in clazzes:
            if clazz in done:
                continue  # Support multiple names of a clazz
            done.append(clazz)
            assert clazz.__name__ != 'DefaultTable', \
                'Oops, table class not found.'
            assert not hasattr(clazz, method.__name__), \
                "Oops, class '%s' has method '%s'." % (clazz.__name__,
                                                       method.__name__)
            setattr(clazz, method.__name__, method)
        return None
    return wrapper


def hintOn(i, hintMaskBytes):
    # used to add the active hints to the bez string,
    # when a T2 hintmask operator is encountered.
    byteIndex = int(i / 8)
    byteValue = hintMaskBytes[byteIndex]
    offset = 7 - (i % 8)
    return ((2**offset) & byteValue) > 0


class T2ToBezExtractor(T2OutlineExtractor):
    # The T2OutlineExtractor class calls a class method as the handler for each
    # T2 operator.
    # I use this to convert the T2 operands and arguments to bez operators.
    # Note: flex is converted to regular rrcurveto's.
    # cntrmasks just map to hint replacement blocks with the specified stems.
    def __init__(self, localSubrs, globalSubrs, nominalWidthX, defaultWidthX,
                 read_hints=True, round_coords=True):
        T2OutlineExtractor.__init__(self, None, localSubrs, globalSubrs,
                                    nominalWidthX, defaultWidthX)
        self.vhints = []
        self.hhints = []
        self.bezProgram = []
        self.read_hints = read_hints
        self.firstMarkingOpSeen = False
        self.closePathSeen = False
        self.subrLevel = 0
        self.round_coords = round_coords
        self.hintMaskBytes = None

    def execute(self, charString):
        self.subrLevel += 1
        SimpleT2Decompiler.execute(self, charString)
        self.subrLevel -= 1
        if (not self.closePathSeen) and (self.subrLevel == 0):
            self.closePath()

    def _point(self, point):
        if self.round_coords:
            return " ".join("%d" % round(pt) for pt in point)
        return " ".join("%3f" % pt for pt in point)

    def rMoveTo(self, point):
        point = self._nextPoint(point)
        if not self.firstMarkingOpSeen:
            self.firstMarkingOpSeen = True
            self.bezProgram.append("sc\n")
        log.debug("moveto %s, curpos %s", point, self.currentPoint)
        self.bezProgram.append("%s mt\n" % self._point(point))
        self.sawMoveTo = True

    def rLineTo(self, point):
        if not self.sawMoveTo:
            self.rMoveTo((0, 0))
        point = self._nextPoint(point)
        log.debug("lineto %s, curpos %s", point, self.currentPoint)
        self.bezProgram.append("%s dt\n" % self._point(point))

    def rCurveTo(self, pt1, pt2, pt3):
        if not self.sawMoveTo:
            self.rMoveTo((0, 0))
        pt1 = list(self._nextPoint(pt1))
        pt2 = list(self._nextPoint(pt2))
        pt3 = list(self._nextPoint(pt3))
        log.debug("curveto %s %s %s, curpos %s", pt1, pt2, pt3,
                  self.currentPoint)
        self.bezProgram.append("%s ct\n" % self._point(pt1 + pt2 + pt3))

    def op_endchar(self, index):
        self.endPath()
        args = self.popallWidth()
        if args:  # It is a 'seac' composite character. Don't process
            raise SEACError

    def endPath(self):
        # In T2 there are no open paths, so always do a closePath when
        # finishing a sub path.
        if self.sawMoveTo:
            log.debug("endPath")
            self.bezProgram.append("cp\n")
        self.sawMoveTo = False

    def closePath(self):
        self.closePathSeen = True
        log.debug("closePath")
        if self.bezProgram and self.bezProgram[-1] != "cp\n":
            self.bezProgram.append("cp\n")
        self.bezProgram.append("ed\n")

    def updateHints(self, args, hint_list, bezCommand):
        self.countHints(args)

        # first hint value is absolute hint coordinate, second is hint width
        if not self.read_hints:
            return

        lastval = args[0]
        arg = str(lastval)
        hint_list.append(arg)
        self.bezProgram.append(arg + " ")

        for i in range(len(args))[1:]:
            val = args[i]
            lastval += val

            if i % 2:
                arg = str(val)
                hint_list.append(arg)
                self.bezProgram.append("%s %s\n" % (arg, bezCommand))
            else:
                arg = str(lastval)
                hint_list.append(arg)
                self.bezProgram.append(arg + " ")

    def op_hstem(self, index):
        args = self.popallWidth()
        self.hhints = []
        self.updateHints(args, self.hhints, "rb")
        log.debug("hstem %s", self.hhints)

    def op_vstem(self, index):
        args = self.popallWidth()
        self.vhints = []
        self.updateHints(args, self.vhints, "ry")
        log.debug("vstem %s", self.vhints)

    def op_hstemhm(self, index):
        args = self.popallWidth()
        self.hhints = []
        self.updateHints(args, self.hhints, "rb")
        log.debug("stemhm %s %s", self.hhints, args)

    def op_vstemhm(self, index):
        args = self.popallWidth()
        self.vhints = []
        self.updateHints(args, self.vhints, "ry")
        log.debug("vstemhm %s %s", self.vhints, args)

    def getCurHints(self, hintMaskBytes):
        curhhints = []
        curvhints = []
        numhhints = len(self.hhints)

        for i in range(int(numhhints / 2)):
            if hintOn(i, hintMaskBytes):
                curhhints.extend(self.hhints[2 * i:2 * i + 2])
        numvhints = len(self.vhints)
        for i in range(int(numvhints / 2)):
            if hintOn(i + int(numhhints / 2), hintMaskBytes):
                curvhints.extend(self.vhints[2 * i:2 * i + 2])
        return curhhints, curvhints

    def doMask(self, index, bezCommand):
        args = []
        if not self.hintMaskBytes:
            args = self.popallWidth()
            if args:
                self.vhints = []
                self.updateHints(args, self.vhints, "ry")
            self.hintMaskBytes = int((self.hintCount + 7) / 8)

        self.hintMaskString, index = self.callingStack[-1].getBytes(
            index, self.hintMaskBytes)

        if self.read_hints:
            curhhints, curvhints = self.getCurHints(self.hintMaskString)
            strout = ""
            mask = [strout + hex(ch) for ch in self.hintMaskString]
            log.debug("%s %s %s %s %s", bezCommand, mask, curhhints, curvhints,
                      args)

            self.bezProgram.append("beginsubr snc\n")
            for i, hint in enumerate(curhhints):
                self.bezProgram.append("%s " % hint)
                if i % 2:
                    self.bezProgram.append("rb\n")
            for i, hint in enumerate(curvhints):
                self.bezProgram.append("%s " % hint)
                if i % 2:
                    self.bezProgram.append("ry\n")
            self.bezProgram.extend(["endsubr enc\n", "newcolors\n"])
        return self.hintMaskString, index

    def op_hintmask(self, index):
        hintMaskString, index = self.doMask(index, "hintmask")
        return hintMaskString, index

    def op_cntrmask(self, index):
        hintMaskString, index = self.doMask(index, "cntrmask")
        return hintMaskString, index

    def countHints(self, args):
        self.hintCount = self.hintCount + int(len(args) / 2)


def convertT2GlyphToBez(t2CharString, read_hints=True, round_coords=True):
    # wrapper for T2ToBezExtractor which
    # applies it to the supplied T2 charstring
    subrs = getattr(t2CharString.private, "Subrs", [])
    extractor = T2ToBezExtractor(subrs,
                                 t2CharString.globalSubrs,
                                 t2CharString.private.nominalWidthX,
                                 t2CharString.private.defaultWidthX,
                                 read_hints,
                                 round_coords)
    extractor.execute(t2CharString)
    t2_width_arg = None
    if extractor.gotWidth and (extractor.width is not None):
        t2_width_arg = extractor.width - t2CharString.private.nominalWidthX
    return "".join(extractor.bezProgram), t2_width_arg


class HintMask:
    # class used to collect hints for the current
    # hint mask when converting bez to T2.
    def __init__(self, listPos):
        # The index into the t2list is kept so we can quickly find them later.
        # Note that t2list has one item per operator, and does not include the
        # initial hint operators - first op is always [rhv]moveto or endchar.
        self.listPos = listPos
        # These contain the actual hint values.
        self.h_list = []
        self.v_list = []
        self.mask = None

    def maskByte(self, hHints, vHints):
        # return hintmask bytes for known hints.
        num_hhints = len(hHints)
        num_vhints = len(vHints)
        self.byteLength = byteLength = int((7 + num_hhints + num_vhints) / 8)
        maskVal = 0
        byteIndex = 0
        mask = b""
        if self.h_list:
            mask, maskVal, byteIndex = self.addMaskBits(
                hHints, self.h_list, 0, mask, maskVal, byteIndex)
        if self.v_list:
            mask, maskVal, byteIndex = self.addMaskBits(
                vHints, self.v_list, num_hhints, mask, maskVal, byteIndex)

        if maskVal:
            mask += bytes([maskVal])

        if len(mask) < byteLength:
            mask += b"\0" * (byteLength - len(mask))
        self.mask = mask
        return mask

    @staticmethod
    def addMaskBits(allHints, maskHints, numPriorHints, mask, maskVal,
                    byteIndex):
        # sort in allhints order.
        sort_list = [[allHints.index(hint) + numPriorHints, hint] for hint in
                     maskHints if hint in allHints]
        if not sort_list:
            # we get here if some hints have been dropped # because of
            # the stack limit, so that none of the items in maskHints are
            # not in allHints
            return mask, maskVal, byteIndex

        sort_list.sort()
        (idx_list, maskHints) = zip(*sort_list)
        for i in idx_list:
            newbyteIndex = int(i / 8)
            if newbyteIndex != byteIndex:
                mask += bytes([maskVal])
                byteIndex += 1
                while byteIndex < newbyteIndex:
                    mask += b"\0"
                    byteIndex += 1
                maskVal = 0
            maskVal += 2**(7 - (i % 8))
        return mask, maskVal, byteIndex

    @property
    def num_bits(self):
        count = sum(
            [bin(mask_byte).count('1') for mask_byte in bytearray(self.mask)])
        return count


def make_hint_list(hints, need_hint_masks, is_h):
    # Add the list of T2 tokens that make up the initial hint operators
    hint_list = []
    lastPos = 0
    # In bez terms, the first coordinate in each pair is absolute,
    # second is relative.
    # In T2, each term is relative to the previous one.
    for hint in hints:
        if not hint:
            continue
        pos1 = hint[0]
        pos = pos1 - lastPos
        if pos % 1 == 0:
            pos = int(pos)
        hint_list.append(pos)
        pos2 = hint[1]
        if pos2 % 1 == 0:
            pos2 = int(pos2)
        lastPos = pos1 + pos2
        hint_list.append(pos2)

    if need_hint_masks:
        if is_h:
            op = "hstemhm"
            hint_list.append(op)
        # never need to append vstemhm: if we are using it, it is followed
        # by a mask command and vstemhm is inferred.
    else:
        if is_h:
            op = "hstem"
        else:
            op = "vstem"
        hint_list.append(op)
    return hint_list


bezToT2 = {
    "mt": 'rmoveto',
    "rmt": 'rmoveto',
    "dt": 'rlineto',
    "ct": 'rrcurveto',
    "cp": '',
    "ed": 'endchar'
}


kHintArgsNoOverlap = 0
kHintArgsOverLap = 1
kHintArgsMatch = 2


def checkStem3ArgsOverlap(arg_list, hint_list):
    status = kHintArgsNoOverlap
    for x0, x1 in arg_list:
        x1 = x0 + x1
        for y0, y1 in hint_list:
            y1 = y0 + y1
            if x0 == y0:
                if x1 == y1:
                    status = kHintArgsMatch
                else:
                    return kHintArgsOverLap
            elif x1 == y1:
                return kHintArgsOverLap
            else:
                if (x0 > y0) and (x0 < y1):
                    return kHintArgsOverLap
                if (x1 > y0) and (x1 < y1):
                    return kHintArgsOverLap
    return status


def _add_cntr_maskHints(counter_mask_list, src_hints, is_h):
    for arg_list in src_hints:
        for mask in counter_mask_list:
            dst_hints = mask.h_list if is_h else mask.v_list
            if not dst_hints:
                dst_hints.extend(arg_list)
                overlap_status = kHintArgsMatch
                break
            overlap_status = checkStem3ArgsOverlap(arg_list, dst_hints)
            # The args match args in this control mask.
            if overlap_status == kHintArgsMatch:
                break
        if overlap_status != kHintArgsMatch:
            mask = HintMask(0)
            counter_mask_list.append(mask)
            dst_hints.extend(arg_list)


def build_counter_mask_list(h_stem3_list, v_stem3_list):

    v_counter_mask = HintMask(0)
    h_counter_mask = v_counter_mask
    counter_mask_list = [h_counter_mask]
    _add_cntr_maskHints(counter_mask_list, h_stem3_list, is_h=True)
    _add_cntr_maskHints(counter_mask_list, v_stem3_list, is_h=False)

    return counter_mask_list


def makeRelativeCTArgs(arg_list, curX, curY):
    newCurX = arg_list[4]
    newCurY = arg_list[5]
    arg_list[5] -= arg_list[3]
    arg_list[4] -= arg_list[2]

    arg_list[3] -= arg_list[1]
    arg_list[2] -= arg_list[0]

    arg_list[0] -= curX
    arg_list[1] -= curY
    return arg_list, newCurX, newCurY


def build_hint_order(hints):
    # MM hints have duplicate hints. We want to return a list of indices into
    # the original unsorted and unfiltered list. The list should be sorted, and
    # should filter out duplicates

    num_hints = len(hints)
    index_list = list(range(num_hints))
    hint_list = list(zip(hints, index_list))
    hint_list.sort()
    new_hints = [hint_list[i] for i in range(1, num_hints)
                 if hint_list[i][0] != hint_list[i - 1][0]]
    new_hints = [hint_list[0]] + new_hints
    hints, hint_order = list(zip(*new_hints))
    # hints is now a list of hint pairs, sorted by increasing bottom edge.
    # hint_order is now a list of the hint indices from the bez file, but
    # sorted in the order of the hint pairs.
    return hints, hint_order


def make_abs(hint_pair):
    bottom_edge, delta = hint_pair
    new_hint_pair = [bottom_edge, delta]
    if delta in [-20, -21]:  # It is a ghost hint!
        # We use this only in comparing overlap and order:
        # pretend the delta is 0, as it isn't a real value.
        new_hint_pair[1] = bottom_edge
    else:
        new_hint_pair[1] = bottom_edge + delta
    return new_hint_pair


def check_hint_overlap(hint_list, last_idx, bad_hint_idxs):
    # return True if there is an overlap.
    prev = hint_list[0]
    for i, hint_pair in enumerate(hint_list[1:], 1):
        if prev[1] >= hint_pair[0]:
            bad_hint_idxs.add(i + last_idx - 1)
        prev = hint_pair


def check_hint_pairs(hint_pairs, mm_hint_info, last_idx=0):
    # pairs must be in ascending order by bottom (or left) edge,
    # and pairs in a hint group must not overlap.

    # check order first
    bad_hint_idxs = set()
    prev = hint_pairs[0]
    for i, hint_pair in enumerate(hint_pairs[1:], 1):
        if prev[0] > hint_pair[0]:
            # If there is a conflict, we drop the previous hint
            bad_hint_idxs.add(i + last_idx - 1)
        prev = hint_pair

    # check for overlap in hint groups.
    if mm_hint_info.hint_masks:
        for hint_mask in mm_hint_info.hint_masks:
            if last_idx == 0:
                hint_list = hint_mask.h_list
            else:
                hint_list = hint_mask.v_list
            hint_list = [make_abs(hint_pair) for hint_pair in hint_list]
            check_hint_overlap(hint_list, last_idx, bad_hint_idxs)
    else:
        hint_list = [make_abs(hint_pair) for hint_pair in hint_pairs]
        check_hint_overlap(hint_list, last_idx, bad_hint_idxs)

    if bad_hint_idxs:
        mm_hint_info.bad_hint_idxs |= bad_hint_idxs


def update_hints(in_mm_hints, arg_list, hints, hint_mask, is_v=False):
    if in_mm_hints:
        hints.append(arg_list)
        i = len(hints) - 1
    else:
        try:
            i = hints.index(arg_list)
        except ValueError:
            i = len(hints)
            hints.append(arg_list)
    if hint_mask:
        hint_list = hint_mask.v_list if is_v else hint_mask.h_list
        if hints[i] not in hint_list:
            hint_list.append(hints[i])
    return i


def convertBezToT2(bezString, mm_hint_info=None):
    # convert bez data to a T2 outline program, a list of operator tokens.
    #
    # Convert all bez ops to simplest T2 equivalent.
    # Add all hints to vertical and horizontal hint lists as encountered.
    # Insert a HintMask class whenever a new set of hints is encountered.
    # Add all hints as prefix to t2Program
    # After all operators have been processed, convert HintMask items into
    # hintmask ops and hintmask bytes.
    # Review operator list to optimize T2 operators.
    #
    # If doing MM-hinting, extra work is needed to maintain merge
    # compatibility between the reference font and the region fonts.
    # Although hints are generated for exactly the same outline features
    # in all fonts, they will have different values. Consequently, the
    # hints in a region font may not sort to the same order as in the
    # reference font. In addition, they may be filtered differently. Only
    # unique hints are added from the bez file to the hint list. Two hint
    # pairs may differ in one font, but not in another.
    # We work around these problems by first not filtering the hint
    # pairs for uniqueness when accumulating the hint lists. For the
    # reference font, once we have collected all the hints, we remove any
    # duplicate pairs, but keep a list of the retained hint pair indices
    # into the unfiltered hint pair list. For the region fonts, we
    # select hints from the unfiltered hint pair lists by using the selected
    # index list from the reference font.
    # Note that this breaks the CFF spec for snapshotted instances of the
    # CFF2 VF variable font, as hints may not be in ascending order, and the
    # hint list may contain duplicate hints.

    in_mm_hints = mm_hint_info is not None
    bezString = re.sub(r"%.+?\n", "", bezString)  # suppress comments
    bezList = re.findall(r"(\S+)", bezString)
    if not bezList:
        return ""
    hhints = []
    vhints = []
    # Always assume a hint mask exists until proven
    # otherwise - make an initial HintMask.
    hint_mask = HintMask(0)
    hintMaskList = [hint_mask]
    vStem3Args = []
    hStem3Args = []
    v_stem3_list = []
    h_stem3_list = []
    arg_list = []
    t2List = []

    lastPathOp = None
    curX = 0
    curY = 0
    for token in bezList:
        try:
            val1 = round(float(token), 2)
            try:
                val2 = int(token)
                if int(val1) == val2:
                    arg_list.append(val2)
                else:
                    arg_list.append("%s 100 div" % int(val1 * 100))
            except ValueError:
                arg_list.append(val1)
            continue
        except ValueError:
            pass

        if token == "newcolors":
            lastPathOp = token
        elif token in ["beginsubr", "endsubr"]:
            lastPathOp = token
        elif token == "snc":
            lastPathOp = token
            # The index into the t2list is kept
            # so we can quickly find them later.
            hint_mask = HintMask(len(t2List))
            t2List.append([hint_mask])
            hintMaskList.append(hint_mask)
        elif token == "enc":
            lastPathOp = token
        elif token == "rb":
            update_hints(in_mm_hints, arg_list, hhints, hint_mask, False)
            arg_list = []
            lastPathOp = token
        elif token == "ry":
            update_hints(in_mm_hints, arg_list, vhints, hint_mask, True)
            arg_list = []
            lastPathOp = token
        elif token == "rm":  # vstem3 hints are vhints
            update_hints(in_mm_hints, arg_list, vhints, hint_mask, True)
            if (lastPathOp != token) and vStem3Args:
                # first rm, must be start of a new vstem3
                # if we already have a set of vstems in vStem3Args, save them,
                # and then clear the vStem3Args so we can add the new set.
                v_stem3_list.append(vStem3Args)
                vStem3Args = []

            vStem3Args.append(arg_list)
            arg_list = []
            lastPathOp = token
        elif token == "rv":  # hstem3 are hhints
            update_hints(in_mm_hints, arg_list, hhints, hint_mask, False)

            if (lastPathOp != token) and hStem3Args:
                # first rv, must be start of a new h countermask
                h_stem3_list.append(hStem3Args)
                hStem3Args = []

            hStem3Args.append(arg_list)
            arg_list = []
            lastPathOp = token
        elif token == "preflx1":
            # The preflx1/preflx2a sequence provides the same 'i' as the flex
            # sequence. The difference is that the preflx1/preflx2a sequence
            # provides the argument values needed for building a Type1 string
            # while the flex sequence is simply the 6 rrcurveto points.
            # Both sequences are always provided.
            lastPathOp = token
            arg_list = []
        elif token == "preflx2a":
            lastPathOp = token
            del t2List[-1]
            arg_list = []
        elif token == "flxa":
            lastPathOp = token
            argList1, curX, curY = makeRelativeCTArgs(arg_list[:6], curX, curY)
            argList2, curX, curY = makeRelativeCTArgs(arg_list[6:], curX, curY)
            arg_list = argList1 + argList2
            t2List.append([arg_list[:12] + [50], "flex"])
            arg_list = []
        elif token == "sc":
            lastPathOp = token
        else:
            if token in ["rmt", "mt", "dt", "ct"]:
                lastPathOp = token
            t2Op = bezToT2.get(token, None)
            if token in ["mt", "dt"]:
                newList = [arg_list[0] - curX, arg_list[1] - curY]
                curX = arg_list[0]
                curY = arg_list[1]
                arg_list = newList
            elif token == "ct":
                arg_list, curX, curY = makeRelativeCTArgs(arg_list, curX, curY)
            if t2Op:
                t2List.append([arg_list, t2Op])
            elif t2Op is None:
                raise KeyError("Unhandled operation %s %s" % (arg_list, token))
            arg_list = []

    # Add hints, if any. Must be done at the end of op processing to make sure
    # we have seen all the hints in the bez string. Note that the hintmask are
    # identified in the t2List by an index into the list; be careful NOT to
    # change the t2List length until the hintmasks have been converted.
    need_hint_masks = len(hintMaskList) > 1
    if vStem3Args:
        v_stem3_list.append(vStem3Args)
    if hStem3Args:
        h_stem3_list.append(hStem3Args)

    t2Program = []

    if hhints or vhints:
        if mm_hint_info is None:
            hhints.sort()
            vhints.sort()
        elif mm_hint_info.defined:
            # Apply hint order from reference font in MM hinting
            hhints = [hhints[j] for j in mm_hint_info.h_order]
            vhints = [vhints[j] for j in mm_hint_info.v_order]
        else:
            # Define hint order from reference font in MM hinting
            hhints, mm_hint_info.h_order = build_hint_order(hhints)
            vhints, mm_hint_info.v_order = build_hint_order(vhints)

        num_hhints = len(hhints)
        num_vhints = len(vhints)
        hint_limit = int((kStackLimit - 2) / 2)
        if num_hhints >= hint_limit:
            hhints = hhints[:hint_limit]
        if num_vhints >= hint_limit:
            vhints = vhints[:hint_limit]

        if mm_hint_info and mm_hint_info.defined:
            check_hint_pairs(hhints, mm_hint_info)
            last_idx = len(hhints)
            check_hint_pairs(vhints, mm_hint_info, last_idx)

        if hhints:
            t2Program = make_hint_list(hhints, need_hint_masks, is_h=True)
        if vhints:
            t2Program += make_hint_list(vhints, need_hint_masks, is_h=False)

        cntrmask_progam = None
        if mm_hint_info is None:
            if v_stem3_list or h_stem3_list:
                counter_mask_list = build_counter_mask_list(h_stem3_list,
                                                            v_stem3_list)
                cntrmask_progam = [['cntrmask', cMask.maskByte(hhints,
                                                               vhints)] for
                                   cMask in counter_mask_list]
        elif (not mm_hint_info.defined):
            if v_stem3_list or h_stem3_list:
                # this is the reference font - we need to build the list.
                counter_mask_list = build_counter_mask_list(h_stem3_list,
                                                            v_stem3_list)
                cntrmask_progam = [['cntrmask', cMask.maskByte(hhints,
                                                               vhints)] for
                                   cMask in counter_mask_list]
                mm_hint_info.cntr_masks = counter_mask_list
        else:
            # This is a region font - we need to used the reference font list.
            counter_mask_list = mm_hint_info.cntr_masks
            cntrmask_progam = [['cntrmask', cMask.mask] for
                               cMask in counter_mask_list]

        if cntrmask_progam:
            cntrmask_progam = itertools.chain(*cntrmask_progam)
            t2Program.extend(cntrmask_progam)

        if need_hint_masks:
            # If there is not a hintsub before any drawing operators, then
            # add an initial first hint mask to the t2Program.
            if (mm_hint_info is None) or (not mm_hint_info.defined):
                # a single font and a reference font for mm hinting are
                # processed the same way
                if hintMaskList[1].listPos != 0:
                    hBytes = hintMaskList[0].maskByte(hhints, vhints)
                    t2Program.extend(["hintmask", hBytes])
                    if in_mm_hints:
                        mm_hint_info.hint_masks.append(hintMaskList[0])

                # Convert the rest of the hint masks
                # to a hintmask op and hintmask bytes.
                for hint_mask in hintMaskList[1:]:
                    pos = hint_mask.listPos
                    hBytes = hint_mask.maskByte(hhints, vhints)
                    t2List[pos] = [["hintmask"], hBytes]
                    if in_mm_hints:
                        mm_hint_info.hint_masks.append(hint_mask)
            elif (mm_hint_info is not None):
                # This is a MM region font:
                # apply hint masks from reference font.
                try:
                    hm0_mask = mm_hint_info.hint_masks[0].mask
                except IndexError:
                    import pdb
                    pdb.set_trace()
                if isinstance(t2List[0][0], HintMask):
                    t2List[0] = [["hintmask"], hm0_mask]
                else:
                    t2Program.extend(["hintmask", hm0_mask])

                for hm in mm_hint_info.hint_masks[1:]:
                    t2List[hm.listPos] = [["hintmask"], hm.mask]

    for entry in t2List:
        try:
            t2Program.extend(entry[0])
            t2Program.append(entry[1])
        except Exception:
            raise KeyError("Failed to extend t2Program with entry %s" % entry)

    if in_mm_hints:
        mm_hint_info.defined = True
    return t2Program


def _run_tx(args):
    try:
        subprocess.check_call(["tx"] + args, stderr=subprocess.DEVNULL)
    except (subprocess.CalledProcessError, OSError) as e:
        raise FontParseError(e)


class FixHintWidthDecompiler(SimpleT2Decompiler):
    # If we are using this class, we know the charstring has hints.
    def __init__(self, localSubrs, globalSubrs, private=None):
        self.hintMaskBytes = 0  # to silence false Codacy error.
        SimpleT2Decompiler.__init__(self, localSubrs, globalSubrs, private)
        self.has_explicit_width = None
        self.h_hint_args = self.v_hint_args = None
        self.last_stem_index = None

    def op_hstem(self, index):
        self.countHints(is_vert=False)
        self.last_stem_index = index
    op_hstemhm = op_hstem

    def op_vstem(self, index):
        self.countHints(is_vert=True)
        self.last_stem_index = index
    op_vstemhm = op_vstem

    def op_hintmask(self, index):
        if not self.hintMaskBytes:
            # Note that I am assuming that there is never an op_vstemhm
            # followed by an op_hintmask. Since this is applied after saving
            # the font with fontTools, this is safe.
            self.countHints(is_vert=True)
            self.hintMaskBytes = (self.hintCount + 7) // 8
        cs = self.callingStack[-1]
        hintMaskBytes, index = cs.getBytes(index, self.hintMaskBytes)
        return hintMaskBytes, index
    op_cntrmask = op_hintmask

    def countHints(self, is_vert):
        args = self.popall()
        if self.has_explicit_width is None:
            if (len(args) % 2) == 0:
                self.has_explicit_width = False
            else:
                self.has_explicit_width = True
                self.width_arg = args[0]
                args = args[1:]
        self.hintCount = self.hintCount + len(args) // 2
        if is_vert:
            self.v_hint_args = args
        else:
            self.h_hint_args = args


class CFFFontData:
    def __init__(self, path, font_format):
        self.inputPath = path
        self.font_format = font_format
        self.mm_hint_info_dict = {}
        self.t2_widths = {}
        self.is_cff2 = False
        self.is_vf = False
        self.vs_data_models = None
        if font_format == "OTF":
            # It is an OTF font, we can process it directly.
            font = TTFont(path)
            if "CFF " in font:
                cff_format = "CFF "
            elif "CFF2" in font:
                cff_format = "CFF2"
                self.is_cff2 = True
            else:
                raise FontParseError("OTF font has no CFF table <%s>." % path)
        else:
            # Else, package it in an OTF font.
            cff_format = "CFF "
            if font_format == "CFF":
                with open(path, "rb") as fp:
                    data = fp.read()
            else:
                fd, temp_path = tempfile.mkstemp()
                os.close(fd)
                try:
                    _run_tx(["-cff", "+b", "-std", path, temp_path])
                    with open(temp_path, "rb") as fp:
                        data = fp.read()
                finally:
                    os.remove(temp_path)

            font = TTFont()
            font['CFF '] = newTable('CFF ')
            font['CFF '].decompile(data, font)

        self.ttFont = font
        self.cffTable = font[cff_format]

        # for identifier in glyph-list:
        # Get charstring.
        self.topDict = self.cffTable.cff.topDictIndex[0]
        self.charStrings = self.topDict.CharStrings
        if 'fvar' in self.ttFont:
            # have not yet collected VF global data.
            self.is_vf = True
            fvar = self.ttFont['fvar']
            CFF2 = self.cffTable
            CFF2.desubroutinize()
            topDict = CFF2.cff.topDictIndex[0]
            # We need a new charstring object into which we can save the
            # hinted CFF2 program data. Copying an existing charstring is a
            # little easier than creating a new one and making sure that all
            # properties are set correctly.
            self.temp_cs = copy.deepcopy(self.charStrings['.notdef'])
            self.vs_data_models = self.get_vs_data_models(topDict,
                                                          fvar)

    def getGlyphList(self):
        return self.ttFont.getGlyphOrder()

    def getPSName(self):
        if self.is_cff2 and 'name' in self.ttFont:
            psName = next((name_rec.string for name_rec in self.ttFont[
                'name'].names if (name_rec.nameID == 6) and (
                    name_rec.platformID == 3)))
            psName = psName.decode('utf-16be')
        else:
            psName = self.cffTable.cff.fontNames[0]
        return psName

    def get_min_max(self, pTopDict, upm):
        if self.is_cff2 and 'hhea' in self.ttFont:
            font_max = self.ttFont['hhea'].ascent
            font_min = self.ttFont['hhea'].descent
        elif hasattr(pTopDict, 'FontBBox'):
            font_max = pTopDict.FontBBox[3]
            font_min = pTopDict.FontBBox[1]
        else:
            font_max = upm * 1.25
            font_min = -upm * 0.25
        alignment_min = min(-upm * 0.25, font_min)
        alignment_max = max(upm * 1.25, font_max)
        return alignment_min, alignment_max

    def convertToBez(self, glyphName, read_hints, round_coords, doAll=False):
        t2Wdth = None
        t2CharString = self.charStrings[glyphName]
        try:
            bezString, t2Wdth = convertT2GlyphToBez(t2CharString,
                                                    read_hints, round_coords)
            # Note: the glyph name is important, as it is used by the C-code
            # for various heuristics, including [hv]stem3 derivation.
            bezString = "% " + glyphName + "\n" + bezString
        except SEACError:
            log.warning("Skipping %s: can't process SEAC composite glyphs.",
                        glyphName)
            bezString = None
        self.t2_widths[glyphName] = t2Wdth
        return bezString

    def updateFromBez(self, bezData, glyphName, mm_hint_info=None):
        t2Program = convertBezToT2(bezData, mm_hint_info)
        if not self.is_cff2:
            t2_width_arg = self.t2_widths[glyphName]
            if t2_width_arg is not None:
                t2Program = [t2_width_arg] + t2Program
        if self.vs_data_models is not None:
            # It is a variable font. Accumulate the charstrings.
            self.glyph_programs.append(t2Program)
        else:
            # This is an MM source font. Update the font's charstring directly.
            t2CharString = self.charStrings[glyphName]
            t2CharString.program = t2Program

    def save(self, path):
        if path is None:
            path = self.inputPath

        if self.font_format == "OTF":
            self.ttFont.save(path)
            self.ttFont.close()
        else:
            data = self.ttFont["CFF "].compile(self.ttFont)
            if self.font_format == "CFF":
                with open(path, "wb") as fp:
                    fp.write(data)
            else:
                fd, temp_path = tempfile.mkstemp()
                os.write(fd, data)
                os.close(fd)

                try:
                    args = ["-t1", "-std"]
                    if self.font_format == "PFB":
                        args.append("-pfb")
                    _run_tx(args + [temp_path, path])
                finally:
                    os.remove(temp_path)

    def close(self):
        self.ttFont.close()

    def isCID(self):
        return hasattr(self.topDict, "FDSelect")

    def hasFDArray(self):
        return self.is_cff2 or hasattr(self.topDict, "FDSelect")

    def flattenBlends(self, blendList):
        if type(blendList[0]) is list:
            flatList = [blendList[i][0] for i in range(len(blendList))]
        else:
            flatList = blendList
        return flatList

    def getFontInfo(self, allow_no_blues, noFlex,
                    vCounterGlyphs, hCounterGlyphs, fdIndex=0):
        # The psautohint library needs the global font hint zones
        # and standard stem widths.
        # Format them into a single text string.
        # The text format is arbitrary, inherited from very old software,
        # but there is no real need to change it.
        pTopDict = self.topDict
        if hasattr(pTopDict, "FDArray"):
            pDict = pTopDict.FDArray[fdIndex]
        else:
            pDict = pTopDict
        privateDict = pDict.Private

        fdDict = fdTools.FDDict()
        fdDict.LanguageGroup = getattr(privateDict, "LanguageGroup", "0")

        if hasattr(pDict, "FontMatrix"):
            fdDict.FontMatrix = pDict.FontMatrix
        else:
            fdDict.FontMatrix = pTopDict.FontMatrix
        upm = int(1 / fdDict.FontMatrix[0])
        fdDict.OrigEmSqUnits = str(upm)

        fdDict.FontName = getattr(pTopDict, "FontName", self.getPSName())

        blueValues = getattr(privateDict, "BlueValues", [])[:]
        numBlueValues = len(blueValues)
        if numBlueValues < 4:
            low, high = self.get_min_max(pTopDict, upm)
            # Make a set of inactive alignment zones: zones outside of the
            # font BBox so as not to affect hinting. Used when source font has
            # no BlueValues or has invalid BlueValues. Some fonts have bad BBox
            # values, so I don't let this be smaller than -upm*0.25, upm*1.25.
            inactiveAlignmentValues = [low, low, high, high]
            if allow_no_blues:
                blueValues = inactiveAlignmentValues
                numBlueValues = len(blueValues)
            else:
                raise FontParseError("Font must have at least four values in "
                                     "its BlueValues array for PSAutoHint to "
                                     "work!")
        blueValues.sort()

        # The first pair only is a bottom zone, where the first value is the
        # overshoot position. The rest are top zones, and second value of the
        # pair is the overshoot position.
        blueValues = self.flattenBlends(blueValues)
        blueValues[0] = blueValues[0] - blueValues[1]
        for i in range(3, numBlueValues, 2):
            blueValues[i] = blueValues[i] - blueValues[i - 1]

        blueValues = [str(v) for v in blueValues]
        numBlueValues = min(numBlueValues, len(fdTools.kBlueValueKeys))
        for i in range(numBlueValues):
            key = fdTools.kBlueValueKeys[i]
            value = blueValues[i]
            setattr(fdDict, key, value)

        if hasattr(privateDict, "OtherBlues"):
            # For all OtherBlues, the pairs are bottom zones, and
            # the first value of each pair is the overshoot position.
            i = 0
            numBlueValues = len(privateDict.OtherBlues)
            blueValues = privateDict.OtherBlues[:]
            blueValues.sort()
            blueValues = self.flattenBlends(blueValues)
            for i in range(0, numBlueValues, 2):
                blueValues[i] = blueValues[i] - blueValues[i + 1]
            blueValues = [str(v) for v in blueValues]
            numBlueValues = min(numBlueValues,
                                len(fdTools.kOtherBlueValueKeys))
            for i in range(numBlueValues):
                key = fdTools.kOtherBlueValueKeys[i]
                value = blueValues[i]
                setattr(fdDict, key, value)

        if hasattr(privateDict, "StemSnapV"):
            vstems = privateDict.StemSnapV
        elif hasattr(privateDict, "StdVW"):
            vstems = [privateDict.StdVW]
        else:
            if allow_no_blues:
                # dummy value. Needs to be larger than any hint will likely be,
                # as the autohint program strips out any hint wider than twice
                # the largest global stem width.
                vstems = [upm]
            else:
                raise FontParseError("Font has neither StemSnapV nor StdVW!")
        vstems.sort()
        vstems = self.flattenBlends(vstems)
        if (len(vstems) == 0) or ((len(vstems) == 1) and (vstems[0] < 1)):
            vstems = [upm]  # dummy value that will allow PyAC to run
            log.warning("There is no value or 0 value for DominantV.")
        fdDict.DominantV = "[" + " ".join([str(v) for v in vstems]) + "]"

        if hasattr(privateDict, "StemSnapH"):
            hstems = privateDict.StemSnapH
        elif hasattr(privateDict, "StdHW"):
            hstems = [privateDict.StdHW]
        else:
            if allow_no_blues:
                # dummy value. Needs to be larger than any hint will likely be,
                # as the autohint program strips out any hint wider than twice
                # the largest global stem width.
                hstems = [upm]
            else:
                raise FontParseError("Font has neither StemSnapH nor StdHW!")
        hstems.sort()
        hstems = self.flattenBlends(hstems)
        if (len(hstems) == 0) or ((len(hstems) == 1) and (hstems[0] < 1)):
            hstems = [upm]  # dummy value that will allow PyAC to run
            log.warning("There is no value or 0 value for DominantH.")
        fdDict.DominantH = "[" + " ".join([str(v) for v in hstems]) + "]"

        if noFlex:
            fdDict.FlexOK = "false"
        else:
            fdDict.FlexOK = "true"

        # Add candidate lists for counter hints, if any.
        if vCounterGlyphs:
            temp = " ".join(vCounterGlyphs)
            fdDict.VCounterChars = "( %s )" % (temp)
        if hCounterGlyphs:
            temp = " ".join(hCounterGlyphs)
            fdDict.HCounterChars = "( %s )" % (temp)

        fdDict.BlueFuzz = getattr(privateDict, "BlueFuzz", 1)

        return fdDict

    def getfdIndex(self, name):
        gid = self.ttFont.getGlyphID(name)
        if hasattr(self.topDict, "FDSelect"):
            fdIndex = self.topDict.FDSelect[gid]
        else:
            fdIndex = 0
        return fdIndex

    def getfdInfo(self, allow_no_blues, noFlex, vCounterGlyphs, hCounterGlyphs,
                  glyphList, fdIndex=0):
        topDict = self.topDict
        fdGlyphDict = None

        # Get the default fontinfo from the font's top dict.
        fdDict = self.getFontInfo(
            allow_no_blues, noFlex, vCounterGlyphs, hCounterGlyphs, fdIndex)
        fontDictList = [fdDict]

        # Check the fontinfo file, and add any other font dicts
        srcFontInfo = os.path.dirname(self.inputPath)
        srcFontInfo = os.path.join(srcFontInfo, "fontinfo")
        if os.path.exists(srcFontInfo):
            with open(srcFontInfo, "r", encoding="utf-8") as fi:
                fontInfoData = fi.read()
            fontInfoData = re.sub(r"#[^\r\n]+", "", fontInfoData)
        else:
            return fdGlyphDict, fontDictList

        if "FDDict" in fontInfoData:
            maxY = topDict.FontBBox[3]
            minY = topDict.FontBBox[1]
            fdGlyphDict, fontDictList, finalFDict = fdTools.parseFontInfoFile(
                fontDictList, fontInfoData, glyphList, maxY, minY,
                self.getPSName())
            if hasattr(topDict, "FDArray"):
                private = topDict.FDArray[fdIndex].Private
            else:
                private = topDict.Private
            if finalFDict is None:
                # If a font dict was not explicitly specified for the
                # output font, use the first user-specified font dict.
                fdTools.mergeFDDicts(fontDictList[1:], private)
            else:
                fdTools.mergeFDDicts([finalFDict], private)
        return fdGlyphDict, fontDictList

    @staticmethod
    def args_to_hints(hint_args):
        hints = [hint_args[0:2]]
        prev = hints[0]
        for i in range(2, len(hint_args), 2):
            bottom = hint_args[i] + prev[0] + prev[1]
            hints.append([bottom, hint_args[i + 1]])
            prev = hints[-1]
        return hints

    @staticmethod
    def extract_hint_args(program):
        width = None
        h_hint_args = []
        v_hint_args = []
        for i, token in enumerate(program):
            if type(token) is str:
                if i % 2 != 0:
                    width = program[0]
                    del program[0]
                    idx = i - 1
                else:
                    idx = i

                if (token[:4] == 'vstem') or token[-3:] == 'mask':
                    h_hint_args = []
                    v_hint_args = program[:idx]

                elif token[:5] == 'hstem':
                    h_hint_args = program[:idx]
                    v_program = program[idx + 1:]

                    for j, vtoken in enumerate(v_program):
                        if type(vtoken) is str:
                            if (vtoken[:5] == 'vstem') or vtoken[-4:] == \
                                    'mask':
                                v_hint_args = v_program[:j]
                                break
                break

        return width, h_hint_args, v_hint_args

    def fix_t2_program_hints(self, program, mm_hint_info, is_reference_font):

        width_arg, h_hint_args, v_hint_args = self.extract_hint_args(program)

        # 1. Build list of good [vh]hints.
        bad_hint_idxs = list(mm_hint_info.bad_hint_idxs)
        bad_hint_idxs.sort()
        num_hhint_pairs = len(h_hint_args) // 2
        for idx in reversed(bad_hint_idxs):
            if idx < num_hhint_pairs:
                hint_args = h_hint_args
                bottom_idx = idx * 2
            else:
                hint_args = v_hint_args
                bottom_idx = (idx - num_hhint_pairs) * 2
            delta = hint_args[bottom_idx] + hint_args[bottom_idx + 1]
            del hint_args[bottom_idx:bottom_idx + 2]
            if len(hint_args) > bottom_idx:
                hint_args[bottom_idx] += delta

        # delete old hints from program
        if mm_hint_info.cntr_masks:
            last_hint_idx = program.index('cntrmask')
        elif mm_hint_info.hint_masks:
            last_hint_idx = program.index('hintmask')
        else:
            for op in ['vstem', 'hstem']:
                try:
                    last_hint_idx = program.index(op)
                    break
                except IndexError:
                    last_hint_idx = None
        if last_hint_idx is not None:
            del program[:last_hint_idx]

        # If there were v_hint_args, but they have now all been
        # deleted, the first token will still be 'vstem[hm]'. Delete it.
        if ((not v_hint_args) and program[0].startswith('vstem')):
            del program[0]

        # Add width and updated hints back.
        if width_arg is not None:
            hint_program = [width_arg]
        else:
            hint_program = []
        if h_hint_args:
            op_hstem = 'hstemhm' if mm_hint_info.hint_masks else 'hstem'
            hint_program.extend(h_hint_args)
            hint_program.append(op_hstem)
        if v_hint_args:
            hint_program.extend(v_hint_args)
            # Don't need to append op_vstem, as this is still in hint_program.
            program = hint_program + program

        # Re-calculate the hint masks.
        if is_reference_font:
            hhints = self.args_to_hints(h_hint_args)
            vhints = self.args_to_hints(v_hint_args)
            for hm in mm_hint_info.hint_masks:
                hm.maskByte(hhints, vhints)

        # Apply fixed hint masks
        if mm_hint_info.hint_masks:
            hm_pos_list = [i for i, token in enumerate(program)
                           if token == 'hintmask']
            for i, hm in enumerate(mm_hint_info.hint_masks):
                pos = hm_pos_list[i]
                program[pos + 1] = hm.mask

        # Now fix the control masks. We will weed out a control mask
        # if it ends up with fewer than 3 hints.
        cntr_masks = mm_hint_info.cntr_masks
        if is_reference_font and cntr_masks:
            # Update mask bytes,
            # and remove control masks with fewer than 3 bits.
            mask_byte_list = [cm.mask for cm in cntr_masks]
            for cm in cntr_masks:
                cm.maskByte(hhints, vhints)
            new_cm_list = [cm for cm in cntr_masks if cm.num_bits >= 3]
            new_mask_byte_list = [cm.mask for cm in new_cm_list]
            if new_mask_byte_list != mask_byte_list:
                mm_hint_info.new_cntr_masks = new_cm_list
        if mm_hint_info.new_cntr_masks:
            # Remove all the old cntrmask ops
            num_old_cm = len(cntr_masks)
            idx = program.index('cntrmask')
            del program[idx:idx + num_old_cm * 2]
            cm_progam = [['cntrmask', cm.mask] for cm in
                         mm_hint_info.new_cntr_masks]
            cm_progam = list(itertools.chain(*cm_progam))
            program[idx:idx] = cm_progam
        return program

    def fix_glyph_hints(self, glyph_name, mm_hint_info,
                        is_reference_font=None):
        # 1. Delete any bad hints.
        # 2. If reference font, recalculate the hint mask byte strings
        # 3. Replace hint masks.
        # 3. Fix cntr masks.
        if self.is_vf:
            # We get called once, and fix all the charstring programs.
            for i, t2_program in enumerate(self.glyph_programs):
                self.glyph_programs[i] = self.fix_t2_program_hints(
                    t2_program, mm_hint_info, is_reference_font=(i == 0))
        else:
            # we are called for each font in turn
            try:
                t2CharString = self.charStrings[glyph_name]
            except KeyError:
                return  # Happens with sparse sources - just skip the glyph.

            program = self.fix_t2_program_hints(t2CharString.program,
                                                mm_hint_info,
                                                is_reference_font)
            t2CharString.program = program

    def get_vf_bez_glyphs(self, glyph_name):
        charstring = self.charStrings[glyph_name]

        if 'vsindex' in charstring.program:
            op_index = charstring.program.index('vsindex')
            vsindex = charstring.program[op_index - 1]
        else:
            vsindex = 0
        self.vsindex = vsindex
        self.glyph_programs = []
        vs_data_model = self.vs_data_model = self.vs_data_models[vsindex]

        bez_list = []
        for vsi in vs_data_model.master_vsi_list:
            t2_program = interpolate_cff2_charstring(charstring, glyph_name,
                                                     vsi.interpolateFromDeltas,
                                                     vsindex)
            self.temp_cs.program = t2_program
            bezString, _ = convertT2GlyphToBez(self.temp_cs, True, True)
            #  DBG Adding glyph name is useful only for debugging.
            bezString = "% {}\n".format(glyph_name) + bezString
            bez_list.append(bezString)
        return bez_list

    @staticmethod
    def get_vs_data_models(topDict, fvar):
        otvs = topDict.VarStore.otVarStore
        region_list = otvs.VarRegionList.Region
        axis_tags = [axis_entry.axisTag for axis_entry in fvar.axes]
        vs_data_models = []
        for vsindex, var_data in enumerate(otvs.VarData):
            vsi = VarStoreInstancer(topDict.VarStore.otVarStore, fvar.axes, {})
            master_vsi_list = [vsi]
            for region_idx in var_data.VarRegionIndex:
                region = region_list[region_idx]
                loc = {}
                for i, axis in enumerate(region.VarRegionAxis):
                    loc[axis_tags[i]] = axis.PeakCoord
                vsi = VarStoreInstancer(topDict.VarStore.otVarStore, fvar.axes,
                                        loc)
                master_vsi_list.append(vsi)
            vdm = VarDataModel(var_data, vsindex, master_vsi_list)
            vs_data_models.append(vdm)
        return vs_data_models

    def merge_hinted_glyphs(self, name):
        new_t2cs = merge_hinted_programs(self.temp_cs, self.glyph_programs,
                                         name, self.vs_data_model)
        if self.vsindex:
            new_t2cs.program = [self.vsindex, 'vsindex'] + new_t2cs.program
        self.charStrings[name] = new_t2cs


def interpolate_cff2_charstring(charstring, gname, interpolateFromDeltas,
                                vsindex):
    # Interpolate charstring
    # e.g replace blend op args with regular args,
    # and discard vsindex op.
    new_program = []
    last_i = 0
    program = charstring.program
    for i, token in enumerate(program):
        if token == 'vsindex':
            if last_i != 0:
                new_program.extend(program[last_i:i - 1])
            last_i = i + 1
        elif token == 'blend':
            num_regions = charstring.getNumRegions(vsindex)
            numMasters = 1 + num_regions
            num_args = program[i - 1]
            # The program list starting at program[i] is now:
            # ..args for following operations
            # num_args values  from the default font
            # num_args tuples, each with numMasters-1 delta values
            # num_blend_args
            # 'blend'
            argi = i - (num_args * numMasters + 1)
            if last_i != argi:
                new_program.extend(program[last_i:argi])
            end_args = tuplei = argi + num_args
            master_args = []
            while argi < end_args:
                next_ti = tuplei + num_regions
                deltas = program[tuplei:next_ti]
                val = interpolateFromDeltas(vsindex, deltas)
                master_val = program[argi]
                master_val += otRound(val)
                master_args.append(master_val)
                tuplei = next_ti
                argi += 1
            new_program.extend(master_args)
            last_i = i + 1
    if last_i != 0:
        new_program.extend(program[last_i:])
    return new_program


def merge_hinted_programs(charstring, t2_programs, gname, vs_data_model):
    num_masters = vs_data_model.num_masters
    var_pen = CFF2CharStringMergePen([], gname, num_masters, 0)
    charstring.outlineExtractor = MergeOutlineExtractor

    for i, t2_program in enumerate(t2_programs):
        var_pen.restart(i)
        charstring.program = t2_program
        charstring.draw(var_pen)

    new_charstring = var_pen.getCharString(
        private=charstring.private,
        globalSubrs=charstring.globalSubrs,
        var_model=vs_data_model, optimize=True)
    return new_charstring


@_add_method(VarStoreInstancer)
def get_scalars(self, vsindex, region_idx):
    varData = self._varData
    # The index key needs to be the master value index, which includes
    # the default font value. VarRegionIndex provides the region indices.
    scalars = {0: 1.0}  # The default font always has a weight of 1.0
    region_index = varData[vsindex].VarRegionIndex
    for idx in range(region_idx):  # omit the scalar for the region.
        scalar = self._getScalar(region_index[idx])
        if scalar:
            scalars[idx + 1] = scalar
    return scalars


class VarDataModel(object):

    def __init__(self, var_data, vsindex, master_vsi_list):
        self.master_vsi_list = master_vsi_list
        self.var_data = var_data
        self._num_masters = len(master_vsi_list)
        self.delta_weights = [{}]  # for default font value
        for region_idx, vsi in enumerate(master_vsi_list[1:]):
            scalars = vsi.get_scalars(vsindex, region_idx)
            self.delta_weights.append(scalars)

    @property
    def num_masters(self):
        return self._num_masters

    def getDeltas(self, master_values, *, round=noRound):
        assert len(master_values) == len(self.delta_weights)
        out = []
        for i, scalars in enumerate(self.delta_weights):
            delta = master_values[i]
            for j, scalar in scalars.items():
                if scalar:
                    delta -= out[j] * scalar
            out.append(round(delta))
        return out



# Copyright 2016 Adobe. All rights reserved.

# Methods:

# Parse args. If glyphlist is from file, read in entire file as single string,
# and remove all white space, then parse out glyph-names and GID's.

# For each font name:
#   Use fontTools library to open font and extract CFF table.
#   If error, skip font and report error.
#   Filter specified glyph list, if any, with list of glyphs in the font.
#   Open font plist file, if any. If not, create empty font plist.
#   Build alignment zone string
#   For identifier in glyph-list:
#     Get T2 charstring for glyph from parent font CFF table. If not present,
#       report and skip.
#     Get new alignment zone string if FDarray index (which font dict is used)
#       has changed.
#     Convert to bez
#     Build autohint point list string; this is used to tell if glyph has been
#       changed since the last time it was hinted.
#     If requested, check against plist dict, and skip if glyph is already
#       hinted or is manually hinted.
#     Call autohint library on bez string.
#     If change to the point list is permitted and happened, rebuild.
#     Autohint point list string.
#     Convert bez string to T2 charstring, and update parent font CFF.
#     Add glyph hint entry to plist file
#  Save font plist file.

import ast
import logging
import os
import sys
import time
from collections import defaultdict, namedtuple

from .otfFont import CFFFontData
from .ufoFont import UFOFontData
from ._psautohint import error as PsAutoHintCError

from . import (get_font_format, hint_bez_glyph, hint_compatible_bez_glyphs,
               FontParseError)

log = logging.getLogger(__name__)


class ACOptions(object):
    def __init__(self):
        self.inputPaths = []
        self.outputPaths = []
        self.reference_font = None
        self.glyphList = []
        self.nameAliases = {}
        self.excludeGlyphList = False
        self.hintAll = False
        self.read_hints = False
        self.allowChanges = False
        self.noFlex = False
        self.noHintSub = False
        self.allow_no_blues = False
        self.hCounterGlyphs = []
        self.vCounterGlyphs = []
        self.logOnly = False
        self.printDefaultFDDict = False
        self.printFDDictList = False
        self.round_coords = True
        self.writeToDefaultLayer = False
        self.baseMaster = {}
        self.font_format = None
        self.report_zones = False
        self.report_stems = False
        self.report_all_stems = False

    def __str__(self):
        # used only when debugging.
        import inspect
        data = []
        methodList = inspect.getmembers(self)
        for fname, fvalue in methodList:
            if fname[0] == "_":
                continue
            data.append(str((fname, fvalue)))
        data.append("")
        return os.linesep.join(data)


class ACHintError(Exception):
    pass


class GlyphReports:
    def __init__(self):
        self.glyphs = {}

    def addGlyphReport(self, glyphName, reportString):
        hstems = {}
        vstems = {}
        hstems_pos = {}
        vstems_pos = {}
        char_zones = {}
        stem_zone_stems = {}
        self.glyphs[glyphName] = [hstems, vstems, char_zones, stem_zone_stems]

        lines = reportString.splitlines()
        for line in lines:
            tokens = line.split()
            key = tokens[0]
            x = ast.literal_eval(tokens[3])
            y = ast.literal_eval(tokens[5])
            hintpos = "%s %s" % (x, y)
            if key == "charZone":
                char_zones[hintpos] = (x, y)
            elif key == "stemZone":
                stem_zone_stems[hintpos] = (x, y)
            elif key == "HStem":
                width = x - y
                # avoid counting duplicates
                if hintpos not in hstems_pos:
                    count = hstems.get(width, 0)
                    hstems[width] = count + 1
                    hstems_pos[hintpos] = width
            elif key == "VStem":
                width = x - y
                # avoid counting duplicates
                if hintpos not in vstems_pos:
                    count = vstems.get(width, 0)
                    vstems[width] = count + 1
                    vstems_pos[hintpos] = width
            else:
                raise FontParseError("Found unknown keyword %s in report file "
                                     "for glyph %s." % (key, glyphName))

    @staticmethod
    def round_value(val):
        if val >= 0:
            return int(val + 0.5)
        else:
            return int(val - 0.5)

    def parse_stem_dict(self, stem_dict):
        """
        stem_dict: {45.5: 1, 47.0: 2}
        """
        # key: stem width
        # value: stem count
        width_dict = defaultdict(int)
        for width, count in stem_dict.items():
            width = self.round_value(width)
            width_dict[width] += count
        return width_dict

    def parse_zone_dicts(self, char_dict, stem_dict):
        all_zones_dict = char_dict.copy()
        all_zones_dict.update(stem_dict)
        # key: zone height
        # value: zone count
        top_dict = defaultdict(int)
        bot_dict = defaultdict(int)
        for top, bot in all_zones_dict.values():
            top = self.round_value(top)
            top_dict[top] += 1
            bot = self.round_value(bot)
            bot_dict[bot] += 1
        return top_dict, bot_dict

    def assemble_rep_list(self, items_dict, count_dict):
        # item 0: stem/zone count
        # item 1: stem width/zone height
        # item 2: list of glyph names
        gorder = list(self.glyphs.keys())
        rep_list = []
        for item in items_dict:
            gnames = list(items_dict[item])
            # sort the names by the font's glyph order
            if len(gnames) > 1:
                gindexes = [gorder.index(gname) for gname in gnames]
                gnames = [x for _, x in sorted(zip(gindexes, gnames))]
            rep_list.append((count_dict[item], item, gnames))
        return rep_list

    def _get_lists(self):
        """
        self.glyphs is a dictionary:
            key: glyph name
            value: list of 4 dictionaries
                   hstems
                   vstems
                   char_zones
                   stem_zone_stems
        {
         'A': [{45.5: 1, 47.0: 2}, {229.0: 1}, {}, {}],
         'B': [{46.0: 2, 46.5: 2, 47.0: 1}, {94.0: 1, 100.0: 1}, {}, {}],
         'C': [{50.0: 2}, {109.0: 1}, {}, {}],
         'D': [{46.0: 1, 46.5: 2, 47.0: 1}, {95.0: 1, 109.0: 1}, {}, {}],
         'E': [{46.5: 2, 47.0: 1, 50.0: 2, 177.0: 1, 178.0: 1},
               {46.0: 1, 75.5: 2, 95.0: 1}, {}, {}],
         'F': [{46.5: 2, 47.0: 1, 50.0: 1, 177.0: 1},
               {46.0: 1, 60.0: 1, 75.5: 1, 95.0: 1}, {}, {}],
         'G': [{43.0: 1, 44.5: 1, 50.0: 1}, {94.0: 1, 109.0: 1}, {}, {}]
        }
        """
        h_stem_items_dict = defaultdict(set)
        h_stem_count_dict = defaultdict(int)
        v_stem_items_dict = defaultdict(set)
        v_stem_count_dict = defaultdict(int)

        top_zone_items_dict = defaultdict(set)
        top_zone_count_dict = defaultdict(int)
        bot_zone_items_dict = defaultdict(set)
        bot_zone_count_dict = defaultdict(int)

        for gName, dicts in self.glyphs.items():
            hStemDict, vStemDict, charZoneDict, stemZoneStemDict = dicts

            glyph_h_stem_dict = self.parse_stem_dict(hStemDict)
            glyph_v_stem_dict = self.parse_stem_dict(vStemDict)

            for stem_width, stem_count in glyph_h_stem_dict.items():
                h_stem_items_dict[stem_width].add(gName)
                h_stem_count_dict[stem_width] += stem_count

            for stem_width, stem_count in glyph_v_stem_dict.items():
                v_stem_items_dict[stem_width].add(gName)
                v_stem_count_dict[stem_width] += stem_count

            glyph_top_zone_dict, glyph_bot_zone_dict = self.parse_zone_dicts(
                charZoneDict, stemZoneStemDict)

            for zone_height, zone_count in glyph_top_zone_dict.items():
                top_zone_items_dict[zone_height].add(gName)
                top_zone_count_dict[zone_height] += zone_count

            for zone_height, zone_count in glyph_bot_zone_dict.items():
                bot_zone_items_dict[zone_height].add(gName)
                bot_zone_count_dict[zone_height] += zone_count

        # item 0: stem count
        # item 1: stem width
        # item 2: list of glyph names
        h_stem_list = self.assemble_rep_list(
            h_stem_items_dict, h_stem_count_dict)

        v_stem_list = self.assemble_rep_list(
            v_stem_items_dict, v_stem_count_dict)

        # item 0: zone count
        # item 1: zone height
        # item 2: list of glyph names
        top_zone_list = self.assemble_rep_list(
            top_zone_items_dict, top_zone_count_dict)

        bot_zone_list = self.assemble_rep_list(
            bot_zone_items_dict, bot_zone_count_dict)

        return h_stem_list, v_stem_list, top_zone_list, bot_zone_list

    @staticmethod
    def _sort_count(t):
        """
        sort by: count (1st item), value (2nd item), list of glyph names (3rd
        item)
        """
        return (-t[0], -t[1], t[2])

    @staticmethod
    def _sort_val(t):
        """
        sort by: value (2nd item), count (1st item), list of glyph names (3rd
        item)
        """
        return (t[1], -t[0], t[2])

    @staticmethod
    def _sort_val_reversed(t):
        """
        sort by: value (2nd item), count (1st item), list of glyph names (3rd
        item)
        """
        return (-t[1], -t[0], t[2])

    def save(self, path):
        h_stems, v_stems, top_zones, bot_zones = self._get_lists()
        items = ([h_stems, self._sort_count],
                 [v_stems, self._sort_count],
                 [top_zones, self._sort_val_reversed],
                 [bot_zones, self._sort_val])
        atime = time.asctime()
        suffixes = (".hstm.txt", ".vstm.txt", ".top.txt", ".bot.txt")
        titles = ("Horizontal Stem List for %s on %s\n" % (path, atime),
                  "Vertical Stem List for %s on %s\n" % (path, atime),
                  "Top Zone List for %s on %s\n" % (path, atime),
                  "Bottom Zone List for %s on %s\n" % (path, atime),
                  )
        headers = (["count    width    glyphs\n"] * 2 +
                   ["count   height    glyphs\n"] * 2)

        for i, item in enumerate(items):
            reps, sortFunc = item
            if not reps:
                continue
            fName = f'{path}{suffixes[i]}'
            title = titles[i]
            header = headers[i]
            with open(fName, "w") as fp:
                fp.write(title)
                fp.write(header)
                reps.sort(key=sortFunc)
                for rep in reps:
                    gnames = ' '.join(rep[2])
                    fp.write(f"{rep[0]:5}    {rep[1]:5}    [{gnames}]\n")
                log.info("Wrote %s" % fName)


def getGlyphID(glyphTag, fontGlyphList):
    if glyphTag in fontGlyphList:
        return fontGlyphList.index(glyphTag)

    return None


def getGlyphNames(glyphTag, fontGlyphList, fontFileName):
    glyphNameList = []
    rangeList = glyphTag.split("-")
    prevGID = getGlyphID(rangeList[0], fontGlyphList)
    if prevGID is None:
        if len(rangeList) > 1:
            log.warning("glyph ID <%s> in range %s from glyph selection "
                        "list option is not in font. <%s>.",
                        rangeList[0], glyphTag, fontFileName)
        else:
            log.warning("glyph ID <%s> from glyph selection list option "
                        "is not in font. <%s>.", rangeList[0], fontFileName)
        return None
    glyphNameList.append(fontGlyphList[prevGID])

    for glyphTag2 in rangeList[1:]:
        gid = getGlyphID(glyphTag2, fontGlyphList)
        if gid is None:
            log.warning("glyph ID <%s> in range %s from glyph selection "
                        "list option is not in font. <%s>.",
                        glyphTag2, glyphTag, fontFileName)
            return None
        for i in range(prevGID + 1, gid + 1):
            glyphNameList.append(fontGlyphList[i])
        prevGID = gid

    return glyphNameList


def filterGlyphList(options, fontGlyphList, fontFileName):
    # Return the list of glyphs which are in the intersection of the argument
    # list and the glyphs in the font.
    # Complain about glyphs in the argument list which are not in the font.
    if not options.glyphList:
        glyphList = fontGlyphList
    else:
        # expand ranges:
        glyphList = []
        for glyphTag in options.glyphList:
            glyphNames = getGlyphNames(glyphTag, fontGlyphList, fontFileName)
            if glyphNames is not None:
                glyphList.extend(glyphNames)
        if options.excludeGlyphList:
            glyphList = [n for n in fontGlyphList if n not in glyphList]
    return glyphList


fontInfoKeywordList = [
    'FontName',  # string
    'OrigEmSqUnits',
    'LanguageGroup',
    'DominantV',  # array
    'DominantH',  # array
    'FlexOK',  # string
    'BlueFuzz',
    'VCounterChars',  # counter
    'HCounterChars',  # counter
    'BaselineYCoord',
    'BaselineOvershoot',
    'CapHeight',
    'CapOvershoot',
    'LcHeight',
    'LcOvershoot',
    'AscenderHeight',
    'AscenderOvershoot',
    'FigHeight',
    'FigOvershoot',
    'Height5',
    'Height5Overshoot',
    'Height6',
    'Height6Overshoot',
    'DescenderOvershoot',
    'DescenderHeight',
    'SuperiorOvershoot',
    'SuperiorBaseline',
    'OrdinalOvershoot',
    'OrdinalBaseline',
    'Baseline5Overshoot',
    'Baseline5',
    'Baseline6Overshoot',
    'Baseline6',
]


def openFile(path, options):
    font_format = get_font_format(path)
    if font_format is None:
        raise FontParseError(f"{path} is not a supported font format")

    if font_format == "UFO":
        font = UFOFontData(path, options.logOnly, options.writeToDefaultLayer)
    else:
        font = CFFFontData(path, font_format)

    return font


def get_glyph_list(options, font, path):
    filename = os.path.basename(path)

    # filter specified list, if any, with font list.
    glyph_list = filterGlyphList(options, font.getGlyphList(), filename)
    if not glyph_list:
        raise FontParseError("Selected glyph list is empty for font <%s>." %
                             filename)

    return glyph_list


def get_bez_glyphs(options, font, glyph_list):
    glyphs = {}

    for name in glyph_list:
        # Convert to bez format
        try:
            bez_glyph = font.convertToBez(name, options.read_hints,
                                          options.round_coords,
                                          options.hintAll)

            if bez_glyph is None or "mt" not in bez_glyph:
                # skip empty glyphs.
                continue
        except KeyError:
            # Source fonts may be sparse, e.g. be a subset of the
            # reference font.
            bez_glyph = None
        glyphs[name] = GlyphEntry(bez_glyph, font)

    total = len(glyph_list)
    processed = len(glyphs)
    if processed != total:
        log.info("Skipped %s of %s glyphs.", total - processed, total)

    return glyphs


def get_fontinfo_list(options, font, glyph_list, is_var=False):

    # Check counter glyphs, if any.
    counter_glyphs = options.hCounterGlyphs + options.vCounterGlyphs
    if counter_glyphs:
        missing = [n for n in counter_glyphs if n not in font.getGlyphList()]
        if missing:
            log.error("H/VCounterChars glyph named in fontinfo is "
                      "not in font: %s", missing)

    # For Type1 name keyed fonts, psautohint supports defining
    # different alignment zones for different glyphs by FontDict
    # entries in the fontinfo file. This is NOT supported for CID
    # or CFF2 fonts, as these have FDArrays, can can truly support
    # different Font.Dict.Private Dicts for different groups of glyphs.
    if font.hasFDArray():
        return get_fontinfo_list_withFDArray(options, font, glyph_list, is_var)
    else:
        return get_fontinfo_list_withFontInfo(options, font, glyph_list)


def get_fontinfo_list_withFDArray(options, font, glyph_list, is_var=False):
    lastFDIndex = None
    fontinfo_list = {}
    for name in glyph_list:
        # get new fontinfo string if FDarray index has changed,
        # as each FontDict has different alignment zones.
        fdIndex = font.getfdIndex(name)
        if not fdIndex == lastFDIndex:
            lastFDIndex = fdIndex
            fddict = font.getFontInfo(options.allow_no_blues,
                                      options.noFlex,
                                      options.vCounterGlyphs,
                                      options.hCounterGlyphs,
                                      fdIndex)
            fontinfo = fddict.getFontInfo()
        fontinfo_list[name] = (fontinfo, None, None)

    return fontinfo_list


def get_fontinfo_list_withFontInfo(options, font, glyph_list):
    # Build alignment zone string
    if options.printDefaultFDDict:
        print("Showing default FDDict Values:")
        fddict = font.getFontInfo(options.allow_no_blues,
                                  options.noFlex,
                                  options.vCounterGlyphs,
                                  options.hCounterGlyphs)
        # Exit by printing default FDDict with all lines indented by one tab
        sys.exit("\t" + "\n\t".join(fddict.getFontInfo().split("\n")))

    fdglyphdict, fontDictList = font.getfdInfo(options.allow_no_blues,
                                               options.noFlex,
                                               options.vCounterGlyphs,
                                               options.hCounterGlyphs,
                                               glyph_list)

    if options.printFDDictList:
        # Print the user defined FontDicts, and exit.
        if fdglyphdict:
            print("Showing user-defined FontDict Values:\n")
            for fi, fontDict in enumerate(fontDictList):
                print(fontDict.DictName)
                print(fontDict.getFontInfo())
                gnameList = []
                # item = [glyphName, [fdIndex, glyphListIndex]]
                itemList = sorted(fdglyphdict.items(), key=lambda x: x[1][1])
                for gName, entry in itemList:
                    if entry[0] == fi:
                        gnameList.append(gName)
                print("%d glyphs:" % len(gnameList))
                if len(gnameList) > 0:
                    gTxt = " ".join(gnameList)
                else:
                    gTxt = "None"
                print(gTxt + "\n")
        else:
            print("There are no user-defined FontDict Values.")
        return

    if fdglyphdict is None:
        fddict = fontDictList[0]
        fontinfo = fddict.getFontInfo()
    else:
        log.info("Using alternate FDDict global values from fontinfo "
                 "file for some glyphs.")

    lastFDIndex = None
    fontinfo_list = {}
    for name in glyph_list:
        if fdglyphdict is not None:
            fdIndex = fdglyphdict[name][0]
            if lastFDIndex != fdIndex:
                lastFDIndex = fdIndex
                fddict = fontDictList[fdIndex]
                fontinfo = fddict.getFontInfo()

        fontinfo_list[name] = (fontinfo, fddict, fdglyphdict)

    return fontinfo_list


class MMHintInfo:
    def __init__(self, glyph_name=None):
        self.defined = False
        self.h_order = None
        self.v_order = None
        self.hint_masks = []
        self.glyph_name = glyph_name
        # bad_hint_idxs contains the hint pair indices for all the bad
        # hint pairs in any of the fonts for the current glyph.
        self.bad_hint_idxs = set()
        self.cntr_masks = []
        self.new_cntr_masks = []
        self.glyph_programs = None

    @property
    def needs_fix(self):
        return len(self.bad_hint_idxs) > 0


def hint_glyph(options, name, bez_glyph, fontinfo):
    try:
        hinted = hint_bez_glyph(fontinfo, bez_glyph, options.allowChanges,
                                not options.noHintSub, options.round_coords,
                                options.report_zones, options.report_stems,
                                options.report_all_stems)
    except PsAutoHintCError:
        raise ACHintError("%s: Failure in processing outline data." %
                          options.nameAliases.get(name, name))

    return hinted


def hint_compatible_glyphs(options, name, bez_glyphs, masters, fontinfo):
    # This function is used by both
    #   hint_with_reference_font->hint_compatible_fonts
    # and hint_vf_font.
    try:
        ref_master = masters[0]
        # *************************************************************
        # *********** DO NOT DELETE THIS COMMENTED-OUT CODE ***********
        # If you're tempted to "clean up", work on solving
        # https://github.com/adobe-type-tools/psautohint/issues/202
        # first, then you can uncomment the "hint_compatible_bez_glyphs"
        # line and remove this and other related comments, as well as
        # the workaround block following "# else:", below. Thanks.
        # *************************************************************
        #
        # if False:
        #     # This is disabled because it causes crashes on the CI servers
        #     # which are not reproducible locally. The branch below is a hack
        #     # to avoid the crash and should be dropped once the crash is
        #     # fixed, https://github.com/adobe-type-tools/psautohint/pull/131
        #     hinted = hint_compatible_bez_glyphs(
        #         fontinfo, bez_glyphs, masters)
        # *** see https://github.com/adobe-type-tools/psautohint/issues/202 ***
        # else:
        hinted = []
        hinted_ref_bez = hint_glyph(options, name, bez_glyphs[0], fontinfo)
        for i, bez in enumerate(bez_glyphs[1:]):
            if bez is None:
                out = [hinted_ref_bez, None]
            else:
                in_bez = [hinted_ref_bez, bez]
                in_masters = [ref_master, masters[i + 1]]
                out = hint_compatible_bez_glyphs(fontinfo,
                                                 in_bez,
                                                 in_masters)
                # out is [hinted_ref_bez, new_hinted_region_bez]
            if i == 0:
                hinted = out
            else:
                hinted.append(out[1])
    except PsAutoHintCError:
        raise ACHintError("%s: Failure in processing outline data." %
                          options.nameAliases.get(name, name))

    return hinted


def get_glyph_reports(options, font, glyph_list, fontinfo_list):
    reports = GlyphReports()

    glyphs = get_bez_glyphs(options, font, glyph_list)
    for name in glyphs:
        if name == ".notdef":
            continue

        bez_glyph = glyphs[name][0]
        fontinfo = fontinfo_list[name][0]

        report = hint_glyph(options, name, bez_glyph, fontinfo)
        reports.addGlyphReport(name, report.strip())

    return reports


GlyphEntry = namedtuple("GlyphEntry", "bez_data,font")


def hint_font(options, font, glyph_list, fontinfo_list):
    aliases = options.nameAliases

    hinted = {}
    glyphs = get_bez_glyphs(options, font, glyph_list)
    for name in glyphs:
        g_entry = glyphs[name]
        fontinfo, fddict, fdglyphdict = fontinfo_list[name]

        if fdglyphdict:
            log.info("%s: Begin hinting (using fdDict %s).",
                     aliases.get(name, name), fddict.DictName)
        else:
            log.info("%s: Begin hinting.", aliases.get(name, name))

        # Call auto-hint library on bez string.
        new_bez_glyph = hint_glyph(options, name, g_entry.bez_data, fontinfo)
        options.baseMaster[name] = new_bez_glyph

        if not ("ry" in new_bez_glyph or "rb" in new_bez_glyph or
                "rm" in new_bez_glyph or "rv" in new_bez_glyph):
            log.info("%s: No hints added!", aliases.get(name, name))
            continue

        if options.logOnly:
            continue

        hinted[name] = GlyphEntry(new_bez_glyph, font)

    return hinted


def hint_compatible_fonts(options, paths, glyphs,
                          fontinfo_list):
    # glyphs is a list of dicts, one per font. Each dict is keyed by glyph name
    # and references a tuple of (src bez file, font)
    aliases = options.nameAliases

    hinted_glyphs = set()
    reference_font = None

    for name in glyphs[0]:
        fontinfo, _, _ = fontinfo_list[name]

        log.info("%s: Begin hinting.", aliases.get(name, name))

        masters = [os.path.basename(path) for path in paths]
        bez_glyphs = [g[name].bez_data for g in glyphs]
        new_bez_glyphs = hint_compatible_glyphs(options, name, bez_glyphs,
                                                masters, fontinfo)
        if options.logOnly:
            continue

        if reference_font is None:
            fonts = [g[name].font for g in glyphs]
            reference_font = fonts[0]
        mm_hint_info = MMHintInfo()

        for i, new_bez_glyph in enumerate(new_bez_glyphs):
            if new_bez_glyph is not None:
                g_entry = glyphs[i][name]
                g_entry.font.updateFromBez(new_bez_glyph, name, mm_hint_info)

        hinted_glyphs.add(name)
        # Now check if we need to fix any hint lists.
        if mm_hint_info.needs_fix:
            reference_font.fix_glyph_hints(name, mm_hint_info,
                                           is_reference_font=True)
            for font in fonts[1:]:
                font.fix_glyph_hints(name,
                                     mm_hint_info,
                                     is_reference_font=False)

    return len(hinted_glyphs) > 0


def hint_vf_font(options, font_path, out_path):
    font = openFile(font_path, options)
    options.noFlex = True  # work around for incompatibel flex args.
    aliases = options.nameAliases
    glyph_names = get_glyph_list(options, font, font_path)
    log.info("Hinting font %s. Start time: %s.", font_path, time.asctime())
    fontinfo_list = get_fontinfo_list(options, font, glyph_names, True)
    hinted_glyphs = set()

    for name in glyph_names:
        fontinfo, _, _ = fontinfo_list[name]
        log.info("%s: Begin hinting.", aliases.get(name, name))

        bez_glyphs = font.get_vf_bez_glyphs(name)
        num_masters = len(bez_glyphs)
        masters = [f"Master-{i}" for i in range(num_masters)]
        new_bez_glyphs = hint_compatible_glyphs(options, name, bez_glyphs,
                                                masters, fontinfo)
        if None in new_bez_glyphs:
            log.info(f"Error while hinting glyph {name}.")
            continue
        if options.logOnly:
            continue
        hinted_glyphs.add(name)

        # First, convert bez to fontTools T2 programs,
        # and check if any hints conflict.
        mm_hint_info = MMHintInfo()
        for i, new_bez_glyph in enumerate(new_bez_glyphs):
            if new_bez_glyph is not None:
                font.updateFromBez(new_bez_glyph, name, mm_hint_info)

        # Now check if we need to fix any hint lists.
        if mm_hint_info.needs_fix:
            font.fix_glyph_hints(name, mm_hint_info)

        # Now merge the programs into a singel CFF2 charstring program
        font.merge_hinted_glyphs(name)

    if hinted_glyphs:
        log.info(f"Saving font file {out_path} with new hints...")
        font.save(out_path)
    else:
        log.info("No glyphs were hinted.")
        font.close()

    log.info("Done with font %s. End time: %s.", font_path, time.asctime())


def hint_with_reference_font(options, fonts, paths, outpaths):
    # We are doing compatible, AKA multiple master, hinting.
    log.info("Start time: %s.", time.asctime())
    options.noFlex = True  # work-around for mm-hinting

    # Get the glyphs and font info of the reference font. We assume the
    # fonts have the same glyph set, glyph dict and in general are
    # compatible. If not bad things will happen.
    glyph_names = get_glyph_list(options, fonts[0], paths[0])
    fontinfo_list = get_fontinfo_list(options, fonts[0], glyph_names)

    glyphs = []
    for i, font in enumerate(fonts):
        glyphs.append(get_bez_glyphs(options, font, glyph_names))

    have_hinted_glyphs = hint_compatible_fonts(options, paths,
                                               glyphs, fontinfo_list)
    if have_hinted_glyphs:
        log.info("Saving font files with new hints...")

        for i, font in enumerate(fonts):
            font.save(outpaths[i])
    else:
        log.info("No glyphs were hinted.")
        font.close()

    log.info("End time: %s.", time.asctime())


def hint_regular_fonts(options, fonts, paths, outpaths):
    # Regular fonts, just iterate over the list and hint each one.
    for i, font in enumerate(fonts):
        path = paths[i]
        outpath = outpaths[i]

        glyph_names = get_glyph_list(options, font, path)
        fontinfo_list = get_fontinfo_list(options, font, glyph_names)

        log.info("Hinting font %s. Start time: %s.", path, time.asctime())

        if options.report_zones or options.report_stems:
            reports = get_glyph_reports(options, font, glyph_names,
                                        fontinfo_list)
            reports.save(outpath)
        else:
            hinted = hint_font(options, font, glyph_names, fontinfo_list)
            if hinted:
                log.info("Saving font file with new hints...")
                for name in hinted:
                    g_entry = hinted[name]
                    font.updateFromBez(g_entry.bez_data, name)
                font.save(outpath)
            else:
                log.info("No glyphs were hinted.")
                font.close()

        log.info("Done with font %s. End time: %s.", path, time.asctime())


def get_outpath(options, font_path, i):
    if options.outputPaths is not None and i < len(options.outputPaths):
        outpath = options.outputPaths[i]
    else:
        outpath = font_path
    return outpath


def hintFiles(options):
    fonts = []
    paths = []
    outpaths = []
    # If there is a reference font, prepend it to font paths.
    # It must be the first font in the list, code below assumes that.
    if options.reference_font:
        font = openFile(options.reference_font, options)
        fonts.append(font)
        paths.append(options.reference_font)
        outpaths.append(options.reference_font)
        if hasattr(font, 'ttFont'):
            assert 'fvar' not in font.ttFont, ("Can't use a CFF2 VF font as a "
                                               "default font in a set of MM "
                                               "fonts.")

    # Open the rest of the fonts and handle output paths.
    for i, path in enumerate(options.inputPaths):
        font = openFile(path, options)
        out_path = get_outpath(options, path, i)
        if hasattr(font, 'ttFont') and 'fvar' in font.ttFont:
            assert not options.report_zones or options.report_stems
            # Certainly not supported now, also I think it only makes sense
            # to ask for zone reports for the source fonts for the VF font.
            # You can't easily change blue values in a VF font.
            hint_vf_font(options, path, out_path)
        else:
            fonts.append(font)
            paths.append(path)
            outpaths.append(out_path)

    if fonts:
        if fonts[0].isCID():
            options.noFlex = True  # Flex hinting in CJK fonts doed bad things.
            # For CFF fonts, being a CID font is a good indicator of being CJK.

        if options.reference_font:
            hint_with_reference_font(options, fonts, paths, outpaths)
        else:
            hint_regular_fonts(options, fonts, paths, outpaths)
